/* $Header: /nw/tony/src/stevie/src/RCS/edit.c,v 1.11 89/08/02 19:57:12 tony Exp $
 *
 * The main edit loop as well as some other simple cursor movement routines.
 */

#include "stevie.h"

/*
 * This flag is used to make auto-indent work right on lines where only
 * a <RETURN> or <ESC> is typed. It is set when an auto-indent is done,
 * and reset when any other editting is done on the line. If an <ESC>
 * or <RETURN> is received, and did_ai is TRUE, the line is truncated.
 */
bool_t	did_ai = FALSE;

void
edit()
{
	void	indentinout();
	extern	bool_t	need_redraw;
	int	c;
	register uchar	*p, *q;
	LPTR	*pos;

	Prenum = 0;

	/* position the display and the cursor at the top of the file. */
	*Topchar = *Filemem;
	*Curschar = *Filemem;
	Cursrow = Curscol = 0;

	do_mlines();		/* check for mode lines before starting */

	updatescreen();

	for ( ;; ) {

	/* Figure out where the cursor is based on Curschar. */
	cursupdate();

	if (need_redraw && !anyinput()) {
		updatescreen();
		need_redraw = FALSE;
	}

	if (!anyinput())
		windgoto(Cursrow,Curscol);


	c = vgetc();

	can_ab = FALSE;

	if (State == NORMAL) {

		/* We're in the normal (non-insert) mode. */

#ifdef KANJI
		if (!isascii(c)) {
			beep();
			continue;
		}
#endif

		/* Pick up any leading digits and compute 'Prenum' */
		if ( (Prenum>0 && isdigit(c)) || (isdigit(c) && c!='0') ){
			Prenum = Prenum*10 + (c-'0');
			continue;
		}
		/* execute the command */
		normal(c);
		Prenum = 0;

	} else {

		/*
		 * Insert or Replace mode.
		 */
		switch (c) {

		case ESC:	/* an escape ends input mode */
		case Ctrl('C'):
#ifdef FEPCTRL
			if (P(P_FC))
				fep_off();
#endif
			/*
			 * If we just did an auto-indent, and have no
			 * trailing part, then truncate the line and
			 * put the cursor back.
			 */
			if (did_ai) {
				if (gchar(Curschar) == NUL) {
					Curschar->linep->s[0] = NUL;
					Curschar->index = 0;
				}
				did_ai = FALSE;
			}
			set_want_col = TRUE;

			/* Don't end up on a '\n' if you can help it. */
			if (gchar(Curschar) == NUL && Curschar->index != 0) {
#ifdef KANJI
				kdec(Curschar);
#else
				dec(Curschar);
#endif
			} else if (Curschar->index != 0) {
				/*
				 * The cursor should end up on the last
				 * inserted character. This is an attempt
				 * to match the real 'vi', but it may not
				 * be quite right yet.
				 */
#ifdef KANJI
				kdec(Curschar);
#else
				dec(Curschar);
#endif
			}

			State = NORMAL;
			msg("");

			/* construct the Redo buffer */
			p=Redobuff;
			q=Insbuff;
			while ( q < Insptr )
				*p++ = *q++;
			*p++ = ESC;
			*p = NUL;
			updatescreen();
			tstpon();
			break;

		case BS:
		case DEL:
			if (!deletechar()) {
				beep();
				break;
			}
			/* cursupdate(); bug in original 3.69 - Junn */
			updateline();
			break;

		case Ctrl('U'):
			if (!deletechar()) {
				beep();
				break;
			}
			while (deletechar())
				;
			updateline();
			break;

		case Ctrl('W'):
			pos = bck_word(Curschar, 0);
			if (!deletechar()) {
				beep();
				break;
			}
			while (Curschar->linep != pos->linep ||
			       Curschar->index > pos->index) {
				if (!deletechar())
					break;
			}
			updateline();
			break;

		case Ctrl('V'):
			c = vgetc();
			if (c != NL) {
				*Insptr++ = Ctrl('V');
				Ninsert++;
				goto ins_one;
			}
			/* fall thru ... */

		case CR:
		case NL:
			can_ab = TRUE;
			*Insptr++ = NL;
			Ninsert++;
			opencmd(FORWARD, TRUE);		/* open a new line */
			break;

		case Ctrl('T'):		/* forward white space */
		case Ctrl('D'):		/* backward white space */
			can_ab = TRUE;
			did_ai = FALSE;
			*Insptr++ = c;
			Ninsert++;
			indentinout(c == Ctrl('D'));
			break;

		default:
			if (0x00 <= c && c < 0x20) {
				beep();
				continue;
			}
		ins_one:
		case TAB:
#ifdef KANJI
			can_ab = ((c & 0xff00) || !isalnum(c) && c != '_');
#else
			can_ab = (!isalnum(c) && c != '_');
#endif
			did_ai = FALSE;
			insertchar(c);
			break;
		}

	}
	}
}

bool_t
deletechar()
{
	/* can't backup past starting point */
	if (Curschar->linep == Insstart->linep &&
	    Curschar->index <= Insstart->index) {
		return FALSE;
	}

	/* can't backup to a previous line */
	if (Curschar->linep != Insstart->linep &&
	    Curschar->index <= 0) {
		return FALSE;
	}

	can_ab = TRUE;
	did_ai = FALSE;
#ifdef KANJI
	kdec(Curschar);
#else
	dec(Curschar);
#endif
	if (State == INSERT)
		delchar(TRUE);
	/*
	 * It's a little strange to put backspaces into
	 * the redo buffer, but it makes auto-indent a
	 * lot easier to deal with.
	 */
	*Insptr++ = BS;
	Ninsert++;
	return TRUE;
}

void
insertchar(c)
int	c;
{
#ifdef KANJI
	if (c & 0xff00) {	/* insert kanji (2 bytes) */
		inschar(c);
		*Insptr++ = (c >> 8) & 0xff;
		*Insptr++ = c & 0xff;
		Ninsert += 2;
	} else {
		inschar(c);
		*Insptr++ = c;
		Ninsert++;
	}
#else
	inschar(c);
	*Insptr++ = c;
	Ninsert++;
#endif

	/*
	 * The following kludge avoids overflowing the statically
	 * allocated insert buffer. Just dump the user back into
	 * command mode, and print a message.
	 */
	if (Insptr+10 >= &Insbuff[IBSIZE]) {
		stuffin(mkstr(ESC));
		emsg("No buffer space - returning to command mode");
		sleep(2);
	}
	updateline();
}

void
getout(havemsg)
bool_t	havemsg;
{
	windgoto(Rows-1,0);
	if (havemsg)
		outstr("\r\n");
	else
		toutstr(T_EL, 1);
	windexit(0);
}

void
scrolldown(nlines)
int	nlines;
{
	register LPTR	*p;
	register int	done = 0;	/* total # of physical lines done */

	/* Scroll up 'nlines' lines. */
	while (nlines--) {
		if ((p = prevline(Topchar)) == NULL)
			break;
		done += plines(p);
		*Topchar = *p;
	}
	/*
	 * If the cursor is on the bottom line, we need to
	 * make sure it gets moved up the appropriate number
	 * of lines so it stays on the screen.
	 */
	if (Curschar->linep == Botchar->linep->prev) {
		int	i = 0;
		while (i < done) {
			i += plines(Curschar);
			*Curschar = *prevline(Curschar);
		}
	}
	s_ins(0, done);
}

void
scrollup(nlines)
int	nlines;
{
	register LPTR	*p;
	register int	done = 0;	/* total # of physical lines done */
	register int	pl;		/* # of plines for the current line */

	/* Scroll down 'nlines' lines. */
	while (nlines--) {
		pl = plines(Topchar);
		if ((p = nextline(Topchar)) == NULL)
			break;
		done += pl;
		if (Curschar->linep == Topchar->linep)
			*Curschar = *p;
		*Topchar = *p;

	}
	s_del(0, done);
}

/*
 * oneright
 * oneleft
 * onedown
 * oneup
 *
 * Move one char {right,left,down,up}.  Return TRUE when
 * sucessful, FALSE when we hit a boundary (of a line, or the file).
 */

bool_t
oneright()
{
	set_want_col = TRUE;

#ifdef KANJI
	switch (kinc(Curschar)) {
	case 0:
		return TRUE;
	case 1:
		kdec(Curschar);		/* crossed a line, so back up */
		/* fall through */
	case -1:
		return FALSE;
	}
	/*NOTREACHED*/
#else
	switch (inc(Curschar)) {
	case 0:
		return TRUE;
	case 1:
		dec(Curschar);		/* crossed a line, so back up */
		/* fall through */
	case -1:
		return FALSE;
	}
	/*NOTREACHED*/
#endif
}

bool_t
oneleft()
{
	set_want_col = TRUE;

#ifdef KANJI
	switch (kdec(Curschar)) {
	case 0:
		return TRUE;
	case 1:
		kinc(Curschar);		/* crossed a line, so back up */
		/* fall through */
	case -1:
		return FALSE;
	}
	/*NOTREACHED*/
#else
	switch (dec(Curschar)) {
	case 0:
		return TRUE;
	case 1:
		inc(Curschar);		/* crossed a line, so back up */
		/* fall through */
	case -1:
		return FALSE;
	}
	/*NOTREACHED*/
#endif
}

void
beginline(flag)
bool_t	flag;
{
#ifdef KANJI
	unsigned int	c;

	while (oneleft())
		;
	if (flag) {
		do {
			c = kgchar(Curschar);
			if (c != ' ' && c != TAB &&
					c != (unsigned int)(KS1<<8|KS2))
				break;
		} while (oneright());
	}
	set_want_col = TRUE;
#else
	while ( oneleft() )
		;
	if (flag) {
		while (isspace(gchar(Curschar)) && oneright())
			;
	}
	set_want_col = TRUE;
#endif
}

bool_t
oneup(n)
int	n;
{
	LPTR	p, *np;
	register int	k;

	p = *Curschar;
	for ( k=0; k<n; k++ ) {
		/* Look for the previous line */
		if ( (np=prevline(&p)) == NULL ) {
			/* If we've at least backed up a little .. */
			if ( k > 0 )
				break;	/* to update the cursor, etc. */
			else
				return FALSE;
		}
		p = *np;
	}
	*Curschar = p;
	/* This makes sure Topchar gets updated so the complete line */
	/* is one the screen. */
	cursupdate();
	/* try to advance to the column we want to be at */
	*Curschar = *coladvance(&p, Curswant);
	return TRUE;
}

bool_t
onedown(n)
int	n;
{
	LPTR	p, *np;
	register int	k;

	p = *Curschar;
	for ( k=0; k<n; k++ ) {
		/* Look for the next line */
		if ( (np=nextline(&p)) == NULL ) {
			if ( k > 0 )
				break;
			else
				return FALSE;
		}
		p = *np;
	}
	/* try to advance to the column we want to be at */
	*Curschar = *coladvance(&p, Curswant);
	return TRUE;
}

/*
 * indentinout(inout)
 *
 * If inout==0, add a white_space to the begining of this line.
 * If inout==1, delete a white_space from the beginning of this line.
 */
static void
indentinout(inout)
int	inout;
{
	register int	white, c;
	int		cindex, sindex;

	cindex = Curschar->index;
	if (Curschar->linep == Insstart->linep) {
		if (Curschar->index > Insstart->index) {
			Curschar->index = Insstart->index;
		}
	}
	else {
		if (Curschar->index > 0) {
			Curschar->index = 0;
		}
	}
	sindex = Curschar->index;
	while (Curschar->index < cindex) {
		if ( ((c = gchar(Curschar)) != TAB) && (c != ' ') ) {
			Curschar->index = cindex;
			return;
		}
		Curschar->index++;
	}

	white = 0;
	Curschar->index = sindex;
	while ( ((c = gchar(Curschar)) == TAB) || (c == ' ') ) {
		delchar(TRUE);
		if (c == TAB) {
			white += P(P_TS);
		}
		else {
			white += 1;
		}
	}
	white -= white % P(P_SW);
	if (inout == 0) {
		white += P(P_SW);
	}
	else {
		white -= P(P_SW);
	}
	while (white >= P(P_TS)) {
		inschar(TAB);
		white -= P(P_TS);
	}
	while (white > 0) {
		inschar(' ');
		white--;
	}
	while ( ((c = gchar(Curschar)) == TAB) || (c == ' ') ) {
		Curschar->index++;
	}
}
