/*
 * The defines in this file establish the environment we're compiling
 * in. Set these appropriately before compiling the editor.
 */

/*
 * One (and only 1) of the following defines should be uncommented.
 * Most of the code is pretty machine-independent. Machine dependent
 * code goes in a file like tos.c or unix.c. The only other place
 * where machine dependent code goes is term.h for escape sequences.
 */
/* #define 	ATARI_MINIX	/* Atari ST + Minix		*/
/* #define 	ATARI_TOS	/* Atari ST + TOS + Sozobon C	*/
/* #define 	UNIX_BSD	/* Unix 4.2BSD or later		*/
/* #define 	UNIX_SYSV	/* Unix System V		*/
/* #define 	OS2		/* OS/2 + MSC5.1		*/
/* #define 	DOS		/* MSDOS + MSC5.1/TC1.5/TC2.0	*/
/* #define 	PC98_VRAM	/* PC-9801 + MSDOS, VRAM direct	*/
/* #define 	PC98_MOUSE	/* PC-9801 + MSDOS, VRAM+MOUSE	*/
/* #define 	PC98_50LINE	/* PC-9801 + MSDOS, 50line mode	*/
/* #define 	PC98_MINIX	/* PC-9801 + Minix1.2		*/

/*
 * Enable or disable kanji handling.
 * One (and only 1) of the following defines should be uncommented.
 */
/* #define 	SJIS		/* Shift-JIS (MS-DOS, OS/2, Unix) */
/* #define 	EUC		/* EUC kanji without kana (Unix) */
/* #define 	JIS		/* JIS kanji without kana (Unix) */
/* #define	BIT8		/* Chars are 8-bit wide, no kanji */
/* #define 	NOKANJI		/* No kanji support */

/*
 * If JIS is defined above, the following must be defined.
 */
#ifdef JIS
#define	KFC	'@'		/* Kanji final character (one of @, B) */
#define	RFC	'J'		/* Roman final character (one of B, J, H) */
#endif

/*
 * The following defines control the inclusion of "optional" features. As
 * the code size of the editor grows, it will probably be useful to be able
 * to tailor the editor to get the features you most want in environments
 * with code size limits.
 *
 * TILDEOP
 *	Normally the '~' command works on a single character. This define
 *	turns on code that allows it to work like an operator. This is
 *	then enabled at runtime with the "tildeop" parameter.
 *
 * HELP
 *	If defined, a series of help screens may be views with the ":help"
 *	command. This eats a fair amount of data space.
 *
 * TERMCAP
 *	Where termcap support is provided, it is generally optional. If
 *	not enabled, you generally get hard-coded escape sequences for
 *	some "reasonable" terminal. In Minix, this means the console. For
 *	UNIX, this means an ANSI standard terminal. See the file "term.h"
 *	for details about specific environments.
 */
#define 	TILDEOP		/* enable tilde to be an operator */
#define 	HELP		/* enable help command */
#define 	TERMCAP		/* enable termcap support */

/*
 * Rather detailed instructions are placed below.
 * In most case, all defines below should be left as they are.
 *
 * KANJI	enable two byte character handling
 * UNIX		general Unix
 * BSD		Berkeley Unix (ioctl, SIGTSTP, setitimer, ...)
 * MINIX	general Minix
 * ATARI	use Atari ST console as a terminal
 * GLOB		perform filename expansion
 * TERMINFO	use terminfo instead of termcap
 * HAS_SYSDIR	has <sys/dir.h> (for UNIX, mostly for BSD)
 * HAS_DIRENT	has <dirent.h> (for UNIX, mostly for SYSV)
 * CPMSTYLE	each line ends with CR/LF, ^Z acts as an EOF
 * V6FORMAT	printf/scanf uses "%D" instead of "%ld"
 * NOBACKQ	backslash behaves like backquote
 * FEPCTRL	Japanese FEP control (for MS-DOS)
 */
#ifndef NOKANJI
#define	KANJI
#endif

#ifdef BIT8
#define	SJIS
#endif

#ifdef ATARI_MINIX
#define	ATARI
#define	MINIX
#endif

#ifdef ATARI_TOS
#define	ATARI
#define	CPMSTYLE
#endif

#ifdef UNIX_BSD
#define	UNIX
#define	BSD
#define	GLOB
#if !defined(HAS_SYSDIR) && !defined(HAS_DIRENT)
#define	HAS_SYSDIR
#endif
#endif

#ifdef UNIX_SYSV
#define	UNIX
#define	GLOB	/* remove this for R2.0 or earlier */
#ifdef TERMCAP
#define	TERMINFO
#endif
#if !defined(HAS_SYSDIR) && !defined(HAS_DIRENT)
#define	HAS_DIRENT
#endif
#endif

#ifdef OS2
#define	GLOB
#define	CPMSTYLE
#define	NOBACKQ
#endif

#ifdef DOS
#define	GLOB
#define	CPMSTYLE
#define	NOBACKQ
#ifdef KANJI
#ifndef BIT8
#define	FEPCTRL
#endif
#endif
#endif

#ifdef PC98_MOUSE
#define	PC98_VRAM
#endif

#ifdef PC98_VRAM
#define	GLOB
#define	CPMSTYLE
#define	NOBACKQ
#ifdef KANJI
#ifndef BIT8
#define	FEPCTRL
#endif
#endif
#ifdef TERMCAP
#undef TERMCAP
#endif
#endif

#ifdef PC98_50LINE
#define	GLOB
#define	CPMSTYLE
#define	NOBACKQ
#ifdef KANJI
#ifndef BIT8
#define	FEPCTRL	/* If this is defined, only 48 lines are effective. */
#endif
#endif
#ifdef TERMCAP
#undef TERMCAP
#endif
#endif

#ifdef PC98_MINIX
#define	MINIX
#define	V6FORMAT
#define	NOBACKQ
#endif
