/*
 * fepctrl.c 1.0 1990/06/06	Public Domain.
 *
 * General purpose Japanese FEP control routines for MS-DOS.
 * Written by Junn Ohta <ohta@src.ricoh.co.jp>
 *
 *	int fep_init()
 *		checks FEP and turn it off, returns FEP type.
 *	void fep_term()
 *		restore the status of FEP saved by fep_init().
 *	void fep_on()
 *		restore the status of FEP saved by fep_off().
 *	void fep_off()
 *		save the status of FEP and turn it off.
 *	void fep_force_on()
 *		turn FEP on by its default "on" status.
 *	void fep_force_off()
 *		don't save the status of FEP and turn it off.
 */

#define 	STEVIE		/* simply comment out this for general use */
#define 	DO_MSKANJI	/* support MS-KANJI application interface */

#ifdef STEVIE
#include "env.h"
#endif

#if !defined(STEVIE) || defined(FEPCTRL)	/* to the end of file */

#include <dos.h>
#include "fepctrl.h"

/*
 * default "on" status of FEP (used only in fep_force_on())
 */
static int fepon[NFEPS] = {
/* FEP_NONE	*/  0,
/* FEP_PC98	*/  1,	/* 0=off, 1=on */
/* FEP_MSKANJI	*/  2,	/* 0=off, 1=system, 2=echo */
/* FEP_VJE	*/  2,	/* 0=off, 1=system, 2=echo */
/* FEP_ATOK6	*/  1,	/* 0=off, 1=roma, 2=kana, 3=alpha, 4=sym, 5=code */
/* FEP_ATOK7	*/  1,	/* 0=off, 1=on */
/* FEP_MTTK86	*/  1,  /* 0=off, 1=on */
/* FEP_MTTK2	*/  1,  /* 0=off, 1=on */
/* FEP_KATANA	*/  1,  /* 0=off, 1=echo, 2=system */
/* FEP_FIXER	*/  1,  /* 0=off, 1=echo, 2=system */
/* FEP_EGB2	*/  2,  /* 0=off, 1=system, 2=echo */
/* FEP_EGB3	*/  2,  /* 0=off, 1=system, 2=echo */
/* FEP_WXP	*/  1,  /* 0=off, 1=echo, 2=system */
/* FEP_MGR2	*/  1,  /* 0=off, 1=on */
/* FEP_NECAIK	*/  1,  /* 0=off, 1=on */
/* FEP_DFJ	*/  1   /* 0=off, 1=on */
};

static int fep = FEP_NONE;
static int oldmode = 0;
static int keepmode = 0;

int
fep_init()
{
	if (fep == FEP_NONE)
		fep = fep_find();
	oldmode = keepmode = fep_mode(0);
	return fep;
}

void
fep_term()
{
	fep_mode(oldmode);
}

void
fep_on()
{
	fep_mode(keepmode);
}

void
fep_off()
{
	keepmode = fep_mode(0);
}

void
fep_force_on()
{
	fep_mode(fepon[fep]);
}

void
fep_force_off()
{
	fep_mode(0);
}

/*--------------------------------------------------------------------*/
#ifdef TEST

static char *fepname[NFEPS] = {
/* FEP_NONE	*/  "(none)",
/* FEP_PC98	*/  "PC9801",
/* FEP_MSKANJI	*/  "MSKANJI",
/* FEP_VJE	*/  "VJE",
/* FEP_ATOK6	*/  "ATOK6",
/* FEP_ATOK7	*/  "ATOK7",
/* FEP_MTTK86	*/  "MTTK86",
/* FEP_MTTK2	*/  "MTTK2",
/* FEP_KATANA	*/  "KATANA",
/* FEP_FIXER	*/  "FIXER",
/* FEP_EGB2	*/  "EGB2",
/* FEP_EGB3	*/  "EGB3",
/* FEP_WXP	*/  "WXP",
/* FEP_MGR2	*/  "MGR2",
/* FEP_NECAIK	*/  "NECAIK",
/* FEP_DFJ	*/  "DFJ"
};

#define	getch()		(bdos(8, 0, 0) & 0xff)
#define	putch(c)	bdos(2, (int)(c), 0)

static void
putstr(s)
char *s;
{
	while (*s)
		putch(*s++);
}

static void
echoline(s)
char *s;
{
	int c;
	putstr(s);
	while ((c = getch()) != '\r' && c != '\n')
		putch(c);
	putstr("\r\n");
}

main()
{
	putstr("fep = ");
	putstr(fepname[fep_init()]);
	putstr("\r\n");
	putstr("enter 4 lines of text\r\n");
	fep_force_on();
	echoline(" on: ");
	fep_off();
	echoline("off: ");
	fep_on();
	echoline(" on: ");
	fep_force_off();
	echoline("off: ");
	fep_term();
	exit(0);
}

#ifdef __TURBOC__
_setargv(){}
_setenvp(){}
exit(c){_exit(c);}
#endif

#endif /* TEST */
/*--------------------------------------------------------------------*/

#ifdef LDATA
#undef LDATA
#endif
#if defined(M_I86CM) || defined(M_I86LM) || defined(M_I86HM)
#define	LDATA
#endif
#if defined(__COMPACT__) || defined(__LARGE__) || defined(__HUGE__)
#define	LDATA
#endif

#define	uns	unsigned

typedef struct {
	int	ftype;		/* fep type */
	char	*dname;		/* device name */
} FEPDEV;

static FEPDEV fepdev[] = {
#ifdef DO_MSKANJI
	{ FEP_MSKANJI,	"MS$KANJI" },
#endif
	{ FEP_VJE,	"AS$VJE"   },
	{ FEP_ATOK6,	"ATOKF%%%" },
	{ FEP_MGR2,	"#MGR#REX" },
	{ FEP_NECAIK,	"$AIC#NEC" },
	{ FEP_NONE, }
};

typedef struct {
	int	ftype;		/* fep type */
	uns int	inum;		/* dointr number */
	uns int	offset;		/* offset of id string */
	char	*idstr;		/* id string */
} FEPINT;

static FEPINT fepint[] = {
	{ FEP_MTTK86,	0x70, 0x12, "MTTK86"   },
	{ FEP_MTTK2,	0x70, 0x12, "MTTK2"    },
	{ FEP_EGB2,	0x70, 0x28, "ERGOA"    },
	{ FEP_EGB3,	0x70, 0x28, "ERGO"     },
	{ FEP_WXP,	0x70, 0x22, "WX      " },
	{ FEP_DFJ,	0x6f, 0x02, "DFJ"      },
/*	{ FEP_VJE,	0x70, 0x12, "VJE-"     }, */
/*	{ FEP_ATOK6,	0x6f, 0x1d, "ATOK5020" }, */
/*	{ FEP_MGR2,	0x70, 0x0a, "#MGR#REX" }, */
/*	{ FEP_NECAIK,	0xdc, 0x0a, "$AIC#NEC" }, */
	{ FEP_NONE, }
};

/*
 * FIXER ioctl data packet
 */
typedef struct {
	char	name[8];
	int	version;
	void	(far *v3func)();	
	int	v3vec;
	void	(far *v4func)();
	int	v4vec;
} FIXIDP;

#ifdef __TURBOC__
static void (far *fixer_entry)();
static void (far *atok7_entry)();
#else
static int fixer_vec = -1;
#endif

#define GET_IOC_DATA	0x4400	/* Get IOCTL Data		*/
#define	RCV_IOC_CHAR	0x4402	/* Receive IOCTL Character	*/
#define	GET_CBRK	0x3300	/* Get CBREAK (CTRL-C Check)	*/
#define	SET_CBRK	0x3301	/* Set CBREAK (CTRL-C Check)	*/
#define	GET_VECTOR	0x35	/* Get Interrupt Vector		*/
#define	OPEN_HANDLE	0x3d	/* Open Handle			*/
#define	CLOSE_HANDLE	0x3e	/* Close Handle			*/
#define DEVICE		0x80	/* handle is not a file		*/

#define	MK_FAR(seg, off) (uns char far *)((uns long)(seg) << 16 | (uns)(off))
#define	FAR_SEG(addr)	 ((unsigned)((unsigned long)(addr) >> 16))
#define	FAR_OFF(addr)	 ((unsigned)(addr))

static union REGS regs;
static struct SREGS sregs;
static unsigned char *p;
static unsigned char far *fp;

#ifndef __TURBOC__	/* Microsoft C */
#define	disable()	_disable()
#define	enable()	_enable()
#endif

static int
fep_find()
{
	int i;
	FEPDEV *fv;
	FEPINT *fi;
	FIXIDP fixer;

	/*
	 * controlled via PC-9801 CRT/KBD BIOS
	 */
	if (check_dev("FP$DUMMY"))
		return FEP_PC98;

	/*
	 * loaded into bank memory by MELCO's FEP loader
	 */
	if (check_dev("FPLD$10")) {
		if (check_dev("AS$VJE"))
			return FEP_VJE;		/* don't use MS-KANJI API */
		if (check_vec(0x70, 0x12, "MTTK2"))
			return FEP_MTTK86;	/* don't use INT 70H I/F */
	}

	/*
	 * determine fep type by its device name
	 */
	for (fv = fepdev; fv->ftype != FEP_NONE; fv++) {
		if (check_dev(fv->dname)) {
#ifdef DO_MSKANJI
			if (fv->ftype != FEP_MSKANJI)
				return fv->ftype;
			if (msknj_find())
				return FEP_MSKANJI;
#else
			return fv->ftype;
#endif
		}
	}

	/*
	 * determine fep type by its id string
	 */
	for (fi = fepint; fi->ftype != FEP_NONE; fi++)
		if (check_vec(fi->inum, fi->offset, fi->idstr))
			return fi->ftype;

#ifdef __TURBOC__
	/*
	 * special case: ATOK7 function call
	 */
	regs.h.ah = GET_VECTOR;
	regs.h.al = 0x6f;
	intdosx(&regs, &regs, &sregs);
	fp = MK_FAR(sregs.es, regs.x.bx + 0x30);
	for (p = "ATOK7"; *p; p++)
		if (*p != *fp++)
			break;
	if (!*p) {
		atok7_entry = (void far *)MK_FAR(sregs.es, regs.x.bx + 0x20);
		return FEP_ATOK7;
	}
#endif

	/*
	 * special case: FIXER3/4 application interface
	 */
	regs.x.ax = GET_IOC_DATA;
	regs.x.bx = 0;	/* stdin */
	intdos(&regs, &regs);
	if (regs.x.dx & DEVICE) {
		regs.x.ax = RCV_IOC_CHAR;
		regs.x.bx = 0;	/* stdin */
		regs.x.cx = sizeof(fixer);
#ifdef LDATA
		regs.x.dx = FAR_OFF(&fixer);
		sregs.ds = FAR_SEG(&fixer);
		intdosx(&regs, &regs, &sregs);
#else
		regs.x.dx = (unsigned int)&fixer;
		intdos(&regs, &regs);
#endif
		i = regs.x.ax;
		if (i >= 8 && !strncmp(fixer.name, "FIXER   ", 8)) {
#ifdef __TURBOC__
			/*
			 * For Turbo C, we always use far call interface.
			 */
			if (i >= 16 && fixer.v3func != (void far *)0) {
				fixer_entry = fixer.v3func;
				return FEP_FIXER;
			}
			if (i >= 22 && fixer.v4func != (void far *)0) {
				fixer_entry = fixer.v4func;
				return FEP_FIXER;
			}
#else /* Microsoft C */
			/*
			 * With Microsoft C, we cannot call the interface
			 * function with some registers set and we have
			 * no way other than using software dointr
			 * interface.
			 */
			if (i >= 16 && fixer.v3vec != -1) {
				fixer_vec = fixer.v3vec;
				return FEP_FIXER;
			}
			if (i >= 22 && fixer.v4vec != -1) {
				fixer_vec = fixer.v4vec;
				return FEP_FIXER;
			}
#endif
		}
	}

	/*
	 * special case: Katana/ACE software dointr interface
	 */
	if (check_vec(0x48, 0x0a, "CON     ") &&
	    check_vec(0x49, 0x0a, "CON     ") &&
	    check_vec(0x4a, 0x0a, "CON     ")) {
		/*
		 * This may not be a good way to check Katana/ACE.
		 * However, console status report (sending "<ESC>@"
		 * returns "$") is not reliable here, because we
		 * may have characters in the keyboard buffer.
		 */
		return FEP_KATANA;
	}

	return FEP_NONE;
}

static int
dopen(dname)
char *dname;
{
	regs.h.ah = OPEN_HANDLE;
	regs.h.al = 0;
#ifdef LDATA
	regs.x.dx = FAR_OFF(dname);
	sregs.ds = FAR_SEG(dname);
	intdosx(&regs, &regs, &sregs);
#else
	regs.x.dx = (unsigned int)dname;
	intdos(&regs, &regs);
#endif
	if (regs.x.cflag)
		return -1;
	return regs.x.ax;
}

static void
dclose(fd)
int fd;
{
	regs.h.ah = CLOSE_HANDLE;
	regs.x.bx = fd;
	intdos(&regs, &regs);
}

static int
check_dev(dname)
char *dname;
{
	int fd, dc;

	if ((fd = dopen(dname)) < 0)
		return 0;

	regs.x.ax = GET_IOC_DATA;
	regs.x.bx = fd;
	intdos(&regs, &regs);
	dc = regs.x.dx;
	dclose(fd);
	return (dc & DEVICE)? 1: 0;
}

static int
check_vec(inum, offset, idstr)
unsigned int inum, offset;
char *idstr;
{
	/*
	 * get dointr vector
	 */
	regs.h.ah = GET_VECTOR;
	regs.h.al = inum;
	intdosx(&regs, &regs, &sregs);
	/*
	 * check id string
	 */
	fp = MK_FAR(sregs.es, offset);
	for (p = idstr; *p; p++)
		if (*p != *fp++)
			return 0;
	return 1;
}

static int
call_fixer(func, val)
int func, val;
{
#ifdef __TURBOC__
	_AX = func;
	_DL = (unsigned char)val;
	(*fixer_entry)();
	return (int)_DL;
#else /* Microsoft C */
	regs.x.ax = func;
	regs.h.dl = (unsigned char)val;
	int86(fixer_vec, &regs, &regs);
	return regs.h.dl;
#endif
}

static int
call_atok7(func, val)
int func, val;
{
#ifdef __TURBOC__
	int	packet[20];

	*packet = val;
	_ES = FP_SEG(packet);
	_BX = FP_OFF(packet);
	_AX = func;
	(*atok7_entry)();
	return *packet;
#else
	return 0;
#endif
}

/*
 * PC-9801 CRT/KBD BIOS
 */
#define	CRTWORK   ((unsigned char far *)MK_FAR(0x0000, 0x053d))
#define	KBDBUF    ((unsigned short far *)MK_FAR(0x0000, 0x0502))
#define	KBDCNT    ((unsigned short far *)MK_FAR(0x0000, 0x0528))
#define	KBDTAIL   ((unsigned short near *(far *))MK_FAR(0x0000, 0x0526))
#define	KBDMAX    16		/* max # of keys in KBDBUF */
#define	CTRLXFER  0xb500	/* CTRL-XFER key scan code */

/*
 * Look into CRT BIOS work area and check FEP status
 */
static int
pc98_crtmode()
{
	return (*CRTWORK == 2)? 1: 0;
}

/*
 * Set key code directly into PC-9801 BIOS key code buffer
 */
static void
pc98_setkey(keycode)
unsigned int keycode;
{
	disable();
	*(unsigned short far *)(MK_FAR(0x0000, *KBDTAIL)) = keycode;
	(*KBDCNT)++;
	(*KBDTAIL)++;
	if (*KBDTAIL >= (unsigned short near *)FAR_OFF(&KBDBUF[KBDMAX]))
		*KBDTAIL = (unsigned short near *)FAR_OFF(KBDBUF);
	enable();
}

static void
pc98_flushkey(keycode)
unsigned int keycode;
{
	while (bdos(6, 0xff, 0) & 0xff) /* void */;
	pc98_setkey(keycode);
	while (bdos(6, 0xff, 0) & 0xff) /* void */;
}

static int
set_cbrk(newcbrk)
int newcbrk;
{
	int curcbrk;

	/*
	 * get current cbrk value
	 */
	regs.x.ax = GET_CBRK;
	intdos(&regs, &regs);
	curcbrk = regs.h.dl;
	/*
	 * set new cbrk value
	 */
	regs.x.ax = SET_CBRK;
	regs.h.dl = newcbrk;
	intdos(&regs, &regs);
	return curcbrk;
}

static int
dointr(vec, ah, al)
int vec, ah, al;
{
	regs.h.ah = ah;
	regs.h.al = al;
	int86(vec, &regs, &regs);
	return regs.h.al;
}

static int
fep_mode(newmode)
int newmode;
{
	int curmode;

	switch (fep) {
#ifdef DO_MSKANJI
	case FEP_MSKANJI:
		/* mode: 0 = off, 1 = system, 2 = echo */
		curmode = msknj_get_mode();
		if (newmode != curmode)
			msknj_set_mode(newmode);
		return curmode;
#endif
	case FEP_VJE:
	case FEP_EGB2:
	case FEP_EGB3:
		/* mode: 0 = off, 1 = system, 2 = echo */
		curmode = dointr(0x70, 0, 0);
		if (newmode != curmode) {
			dointr(0x70, 1, newmode);
			if (fep == FEP_EGB2)
				dointr(0x70, 12, 0); /* unlock mode key */
		}
		return curmode;
	case FEP_ATOK6:
		/* mode: 0=off, 1=roma, 2=kana, 3=alpha, 4=sym, 5=code */
		curmode = dointr(0x6f, 102, 0);
		if (newmode != curmode)
			dointr(0x6f, (!newmode? 11: newmode), 0);
		return curmode;
	case FEP_ATOK7:
		/* mode: 0 = off, 1 = on */
		curmode = call_atok7(3, -1)? 1: 0;
		if (newmode != curmode)
			call_atok7(newmode? 1: 2, 0);
		return curmode;
	case FEP_MTTK2:
		/* mode: 0 = off, 1 = on */
		regs.h.ah = 21;
		int86x(0x70, &regs, &regs, &sregs);
		fp = MK_FAR(sregs.es, regs.x.bx);
		curmode = fp[5];
		if (newmode != curmode)
			dointr(0x70, 0, newmode);
		return curmode;
	case FEP_FIXER:
		/* mode: 0 = off, 1 = echo, 2 = system */
		curmode = call_fixer(0x0901, 0);
		if (newmode != curmode)
			call_fixer(0x0900, newmode);
		return curmode;
	case FEP_KATANA:
		/* mode: 0 = off, 1 = echo, 2 = system */
		curmode = dointr(0x48, 2, 0);
		if (newmode != curmode) {
			int old = set_cbrk(0);
			dointr(0x48, 3, newmode);
			set_cbrk(old);
		}
		return curmode;
	case FEP_WXP:
		/* mode: 0 = off, 1 = echo, 2 = system */
		curmode = dointr(0x70, 0x4c, 0);
		if (newmode == 0 && curmode != 0) {
			while (dointr(0x70, 0x4c, 0))
				pc98_flushkey(CTRLXFER);
		} else if (newmode != 0 && curmode == 0) {
			dointr(0x70, 0x5c, newmode);
			dointr(0x70, 0x52, 1);
		}
		return curmode;
	case FEP_MGR2:
		/* mode: 0 = off, 1 = on */
		regs.h.ah = 0;
		int86(0x70, &regs, &regs);
		curmode = (regs.h.dl & 0x01);
		if (newmode != curmode)
			dointr(0x46, 0, newmode);
		return curmode;
	case FEP_MTTK86:
	case FEP_DFJ:
		/* mode: 0 = off, 1 = on */
		curmode = pc98_crtmode();
		while (pc98_crtmode() != newmode)
			pc98_flushkey(CTRLXFER);
		return curmode;
	case FEP_PC98:
	case FEP_NECAIK:
		/* mode: 0 = off, 1 = on */
		curmode = pc98_crtmode();
		if (newmode != curmode)
			pc98_setkey(CTRLXFER);
		return curmode;
	default:
		return 0;
	}
}

#ifdef DO_MSKANJI
/*
 * MS-KANJI Application Interface access routines
 */

struct Kkname {
	short		wLevel;			/* API support level */
	char		rgchName[8];		/* name of system */
	char		rgchVersion[4];		/* version # of system */
	char		rgchRevision[4];	/* revision # of system */
	char		Reserved[14];		/* reserved */
};
struct Dataparm {
	short		wType;			/* display switch */
	short		wScan;			/* key scan code */
	short		wAscii;			/* ascii code */
	short		wShift;			/* shift key status */
	short		wExShift;		/* extended shift key status */
	short		cchResult;		/* size of confirmed string */
	char far	*lpchResult;		/* ptr to confirmed string */
	short		cchMode;		/* size of mode display str */
	char far	*lpchMode;		/* ptr to mode display str */
	char far	*lpattrMode;		/* ptr to attribute of it */
	short		cchSystem;		/* size of system string */
	char far	*lpchSystem;		/* ptr to system string */
	char far	*lpattrSystem;		/* ptr to attribute of it */
	short		ccBuf;			/* size of unconfirmed str */
	char far	*lpchBuf;		/* ptr to unconfirmed str */
	char far	*lpattrBuf;		/* ptr to attribute of it */
	short		cchBufCursor;		/* cursor position in it */
	char		Reserved[34];		/* reserved */
};
struct Funcparm {
	short		wFunc;			/* function # */
	short		wMode;			/* reference or set */
	struct Kkname far *lpKkname;		/* translation system */
	struct Dataparm far *lpDataparm;	/* data parameter */
	char		Reserved[4];		/* reserved */
};

/*
 * function # for wFunc
 */
#define	KKAsk	1
#define	KKOpen	2
#define KKClose	3
#define	KKInOut	4
#define	KKMode	5

/*
 * mode bit for wMode
 */
#define	KM_GET	0x0000
#define	KM_SET	0x8000
#define	KM_OFF	0x0001
#define	KM_ON	0x0002
#define	KM_SYS	0x0004
#define	KM_ECHO	0x0008

static int (far pascal *KKfunc)();
static int (far *KKfunc_vje)();
static struct Funcparm Funcbuf;
static struct Kkname Kkbuf;

static int
msknj_find()
{
	int fd, n;

	/*
	 * get entry address of KKfunc()
	 */
	fd = dopen("MS$KANJI");
	regs.x.ax = RCV_IOC_CHAR;
	regs.x.bx = fd;
	regs.x.cx = 4;
#ifdef LDATA
	regs.x.dx = FAR_OFF(&KKfunc);
	sregs.ds = FAR_SEG(&KKfunc);
	intdosx(&regs, &regs, &sregs);
#else
	regs.x.dx = (unsigned int)&KKfunc;
	intdos(&regs, &regs);
#endif
	n = regs.x.ax;
	dclose(fd);
	if (n != 4)		/* ioctl failed */
		return 0;

	/*
	 * VJE-Beta's KKfunc doesn't restore the stack pointer,
	 * so we should call KKfunc() as a C function instead
	 * of a pascal one.  (What is the *standard* ?)
	 */
	if (check_dev("AS$VJE"))
		KKfunc_vje = (int (far *)())KKfunc;

	/*
	 * check API support level
	 */
	memset(&Funcbuf, 0, sizeof(Funcbuf));
	Funcbuf.wFunc = KKAsk;
	Funcbuf.lpKkname = (struct Kkname far *)&Kkbuf;
	if (msknj_kkfunc((struct Funcparm far *)&Funcbuf) < 0)
		return 0;
	if (Kkbuf.wLevel < 1)	/* level too low */
		return 0;
	return 1;
}

static int
msknj_get_mode()
{
	memset(&Funcbuf, 0, sizeof(Funcbuf));
	Funcbuf.wFunc = KKMode;
	Funcbuf.wMode = KM_GET;
	msknj_kkfunc((struct Funcparm far *)&Funcbuf);
	return (Funcbuf.wMode & KM_OFF)? 0: (Funcbuf.wMode & KM_SYS)? 1: 2;
}

static int
msknj_set_mode(mode)
int mode;
{
	memset(&Funcbuf, 0, sizeof(Funcbuf));
	Funcbuf.wFunc = KKMode;
	Funcbuf.wMode = KM_SET;
	Funcbuf.wMode |= (mode == 0)? KM_OFF: KM_ON;
	Funcbuf.wMode |= (mode == 1)? KM_SYS: KM_ECHO;
	msknj_kkfunc((struct Funcparm far *)&Funcbuf);
	return 0;
}

static int
msknj_kkfunc(bufp)
struct Funcparm far *bufp;
{
	if (KKfunc_vje)
		return (*KKfunc_vje)((struct Funcparm far *)bufp);
	else
		return (*KKfunc)((struct Funcparm far *)bufp);
}
#endif /* DO_MSKANJI */

#endif /* !STEVIE || FEPCTRL */
