/* $Header: /nw/tony/src/stevie/src/RCS/fileio.c,v 1.12 89/08/06 09:50:01 tony Exp $
 *
 * Basic file I/O routines.
 */

#include "stevie.h"

void
filemess(s)
uchar	*s;
{
	smsg("\"%s\" %s", (Filename == NULL) ? (uchar *)"" : Filename, s);
	flushbuf();
}

void
renum()
{
	LPTR	*p;
	unsigned long l = 0;

	for (p = Filemem; p != NULL ;p = nextline(p), l += LINEINC)
		p->linep->num = l;

	Fileend->linep->num = 0xffffffff;
}

bool_t
readfile(fname,fromp,nochangename)
uchar	*fname;
LPTR	*fromp;
bool_t	nochangename;	/* if TRUE, don't change the Filename */
{
	FILE	*f, *fopen();
	register LINE	*curr;
	uchar	buff[MAXLINE], buf2[MAXMLEN];
	register int	i, c;
	register long	nchars = 0;
	int	linecnt = 0;
	bool_t	wasempty = bufempty();
	int	nulls = 0;		/* count nulls */
	bool_t	incomplete = FALSE;	/* was the last line incomplete? */
	bool_t	toolong = FALSE;	/* a line was too long */
	bool_t	nomem = FALSE;		/* no more memory */
	bool_t	readonly = FALSE;	/* file is readonly */
	int	st;			/* file access status */
#ifdef KANJI
#ifdef JIS
	bool_t	kmode = FALSE;		/* in kanji mode? */
	int	csave = NUL;
#endif
	int	badkanji = 0;		/* count orphan kanjis */
	int	kstate = 0;		/* 0: out of kanji */
					/* 1: 1st byte of kanji */
					/* 2: 2nd byte of kanji */
#else
	int	nonascii = 0;		/* count garbage characters */
#endif

	curr = fromp->linep;

	if (!nochangename) {
		readonly = P(P_RO);
		if ((Filename = strsave(fname)) == NULL)
			return FALSE;
		st = filestat(fname);
		if ((st & FS_EXIST) == 0)
			return TRUE;
		if ((st & FS_READ) == 0) {
			emsg("Permission denied");
			return FALSE;
		}
		if ((st & FS_WRITE) == 0)
			readonly = TRUE;
		if (st & FS_DIR) {
			filemess("Directory");
			return FALSE;
		}
	}

	if ((f = fopenb(fixname(fname), "r")) == NULL)
		return TRUE;

	filemess("");

	i = 0;
	do {
#ifdef KANJI
#ifdef JIS
		if (csave) {
			c = csave;
			csave = NUL;
		} else {
		again:
			c = getc(f);
			if (c == ESC) {
				c = getc(f);
				if (c == '$' || c == '(') {
					getc(f);
					kmode = (c == '$');
					goto again;
				}
				csave = c;
				c = ESC;
			}
		}
		if (kmode && isKanji(c | 0x80))
			c |= 0x80;
		kstate = (kstate == 1)? 2: (isKanji(c))? 1: 0;
#else
		c = getc(f);
		kstate = (kstate == 1)? 2: (isKanji(c))? 1: 0;
#endif
#else /*KANJI*/
		c = getc(f);
#endif /*KANJI*/

		if (c == EOF) {
			if (i == 0)	/* normal loop termination */
				break;

#ifdef CPMSTYLE
			if (i == 1 && buff[0] == Ctrl('Z'))
				/*
				 * got a CP/M style EOF;
				 * gracefully discard it
				 */
				break;
#endif

			/*
			 * If we get EOF in the middle of a line, note the
			 * fact and complete the line ourselves.
			 */
			incomplete = TRUE;
#ifdef KANJI
			if (kstate == 2) {
				i--;
				badkanji++;
			}
			kstate = 0;
#endif
			c = NL;
		}

		/*
		 * Abort if we get an interrupt, but finished reading the
		 * current line first.
		 */
		if (got_int && i == 0)
			break;

#ifndef KANJI
		if (c >= 0x80) {
			c -= 0x80;
			nonascii++;
		}
#endif

		/*
		 * If we reached the end of the line, OR we ran out of
		 * space for it, then process the complete line.
		 */
		if (c == NL || i == (MAXLINE-1)) {
			LINE	*lp;

#ifdef KANJI
			if (c != NL) {
				i -= kstate;
				toolong = TRUE;
			} else if (kstate == 2) {
				i--;
				badkanji++;
			}
			kstate = 0;
#else
			if (c != NL)
				toolong = TRUE;
#endif

			if (c == NL && P(P_CR) && buff[i-1] == CR)
				i--;
			buff[i] = NUL;
			if ((lp = newline(i)) == NULL) {
				nomem = TRUE;
				break;
			}

			strcpy(lp->s, buff);

			curr->next->prev = lp;	/* new line to next one */
			lp->next = curr->next;

			curr->next = lp;	/* new line to prior one */
			lp->prev = curr;

			curr = lp;		/* new line becomes current */
			i = 0;
			linecnt++;

		} else if (c == NUL)
			nulls++;		/* count and ignore nulls */
		else {
			buff[i++] = c;		/* normal character */
		}

		nchars++;

	} while (!incomplete && !toolong);

	fclose(f);

	/*
	 * If the buffer was empty when we started, we have to go back
	 * and remove the "dummy" line at Filemem and patch up the ptrs.
	 */
	if (wasempty && nchars != 0) {
		LINE	*dummy = Filemem->linep;	/* dummy line ptr */
		LPTR	*mark = getmark('\'');
		bool_t	unmarked = (mark && mark->linep == dummy);

		free(dummy->s);				/* free string space */
		Filemem->linep = Filemem->linep->next;
		free((char *)dummy);			/* free LINE struct */
		Filemem->linep->prev = Filetop->linep;
		Filetop->linep->next = Filemem->linep;

		Curschar->linep = Filemem->linep;
		Topchar->linep  = Filemem->linep;

		if (unmarked)
			setpcmark();
	}

	renum();

	if (nomem) {
		return FALSE;
	}

	if (got_int) {
		smsg("\"%s\" Interrupt", fname);
		got_int = FALSE;
		return FALSE;		/* an interrupt isn't really an error */
	}

	if (toolong) {
		smsg("\"%s\" Line too long", fname);
		return FALSE;
	}

#ifdef V6FORMAT
	sprintf(buff, "\"%s\" %s%s%d line%s, %D character%s",
#else
	sprintf(buff, "\"%s\" %s%s%d line%s, %ld character%s",
#endif
		fname,
		readonly ? "[Read only] " : "",
		incomplete ? "[Incomplete last line] " : "",
		linecnt, (linecnt != 1) ? "s" : "",
		nchars, (nchars != 1) ? "s" : "");

	buf2[0] = NUL;

#ifdef KANJI
	if (nulls || badkanji) {
		strcat(buff, " (");
		if (nulls) {
			sprintf(buf2, "%d null", nulls);
			strcat(buff, buf2);
		}
		if (badkanji) {
			if (nulls)
				strcat(buff, ", ");
			sprintf(buf2, "%d bad-KANJI", badkanji);
			strcat(buff, buf2);
		}
		strcat(buff, ")");
	}
#else
	if (nonascii || nulls) {
		if (nonascii) {
			if (nulls)
				sprintf(buf2, " (%d null, %d non-ASCII)",
					nulls, nonascii);
			else
				sprintf(buf2, " (%d non-ASCII)", nonascii);
		} else
			sprintf(buf2, " (%d null)", nulls);
	}
	strcat(buff, buf2);
#endif
	msg(buff);

	return FALSE;
}


/*
 * writeit - write to file 'fname' lines 'start' through 'end'
 *
 * If either 'start' or 'end' contain null line pointers, the default
 * is to use the start or end of the file respectively.
 */
bool_t
writeit(fname, start, end, force)
uchar	*fname;
LPTR	*start, *end;
bool_t	force;
{
	FILE	*f, *fopen();
	register uchar	*s;
	register long	nchars;
	register int	lines;
	register LPTR	*p;
	int	st;
	bool_t	nospace = FALSE;
#ifdef JIS
	bool_t	kmode = FALSE;
#endif

	st = filestat(fname);
	if (st & FS_DIR) {
		emsg("Is a directory");
		return FALSE;
	}
	if (!force) {
		if (fname != Filename && (st & FS_EXIST) != 0) {
			emsg("File exists - use \"w!\" to overwrite");
			return FALSE;
		}
		if (fname == Filename && (st & FS_EXIST) != 0 &&
				(P(P_RO) || (st & FS_WRITE) == 0)) {
			emsg("File is read only");
			return FALSE;
		}
	}

	smsg("\"%s\"", fname);

	/*
	 * If 'backup' option is set and the current version exists,
	 * delete any existing backup and move the current version
	 * to the backup.
	 */
	if (P(P_BK) && (st & FS_EXIST) != 0) {
		if (Rename(fname, backupname(fname, ".bak")) == -1) {
			emsg("Permission denied (can't backup)");
			return FALSE;
		}
	}

	f = fopenb(fixname(fname), "w");

	if (f == NULL) {
		emsg("Permission denied");
		return FALSE;
	}

	/*
	 * If we were given a bound, start there. Otherwise just
	 * start at the beginning of the file.
	 */
	if (start == NULL || start->linep == NULL)
		start = Filemem;

	p = start;
	lines = nchars = 0;
	do {
#ifdef JIS
		s = p->linep->s;
		for (;;) {
			if (kmode && !(*s & 0x80)) {
				putc(ESC, f);
				putc('(', f);
				putc(RFC, f);
				nchars += 3;
				kmode = FALSE;
			} else if (!kmode && (*s & 0x80)) {
				putc(ESC, f);
				putc('$', f);
				putc(KFC, f);
				nchars += 3;
				kmode = TRUE;
			}
			if (*s) {
				putc((*s++ & 0x7f), f);
				nchars++;
				continue;
			}
			if (P(P_CR))
				putc(CR, f);
			putc(NL, f);
			nchars++;
			break;
		}
#else
		fprintf(f, "%s%s", p->linep->s, P(P_CR)? "\r\n": "\n");
		nchars += strlen(p->linep->s) + 1;
#endif
		lines++;

		if (ferror(f)) {
			nospace = TRUE;
			break;
		}

		/*
		 * If we were given an upper bound, and we just did that
		 * line, then bag it now.
		 */
		if (end != NULL && end->linep != NULL) {
			if (end->linep == p->linep)
				break;
		}

	} while ((p = nextline(p)) != NULL);

	fclose(f);

	if (nospace) {
		uchar	buf[MAXMLEN];
		sprintf(buf,
		"No space left on device [Warning - %s is incomplete]", fname);
		emsg(buf);
		return FALSE;
	}

#ifdef V6FORMAT
	smsg("\"%s\" %d line%s, %D character%s", fname,
#else
	smsg("\"%s\" %d line%s, %ld character%s", fname,
#endif
		lines, (lines > 1) ? "s" : "",
		nchars, (nchars > 1) ? "s" : "");

	/*
	 * We assume that all modifications are saved
	 * if the entire buffer is written into a file
	 */
	if (prevline(start) == NULL &&
	    (end == NULL || end->linep == NULL || nextline(end) == NULL))
		UNCHANGED;

	return TRUE;
}
