/* $Header: /nw/tony/src/stevie/src/RCS/ops.c,v 1.5 89/08/06 09:50:42 tony Exp $
 *
 * Contains routines that implement the operators in vi. Everything in this
 * file is called only from code in normal.c
 */

#include "stevie.h"
#include "ops.h"

/*
 * doshift - handle a shift operation
 */
void
doshift(op, c1, c2, num)
int	op;
#ifdef KANJI
unsigned int	c1, c2;
#else
uchar	c1, c2;
#endif
int	num;
{
	void	whiteinout();
	LPTR	top, bot;
	int	nlines;
	uchar	opchar;
	register int	c;

	top = startop;
	bot = *Curschar;

	if (lt(&bot, &top))
		pswap(&top, &bot);

	u_save(top.linep->prev, bot.linep->next);

	nlines = cntllines(&top, &bot);
	*Curschar = top;
	whiteinout((op == LSHIFT), nlines);

	/* construct Redo buff */
	opchar = (op == LSHIFT) ? '<' : '>';
#ifdef KANJI
	if (num != 0)
		sprintf(Redobuff, "%c%d%c%s", opchar, num, c1, mkstr(c2));
	else
		sprintf(Redobuff, "%c%c%s", opchar, c1, mkstr(c2));
#else
	if (num != 0)
		sprintf(Redobuff, "%c%d%c%c", opchar, num, c1, c2);
	else
		sprintf(Redobuff, "%c%c%c", opchar, c1, c2);
#endif

	/*
	 * The cursor position afterward is the prior of the two positions.
	 */
	*Curschar = top;
	Curschar->index = 0;
	while ( ((c = gchar(Curschar)) == ' ') || (c == TAB) ) {
		Curschar->index++;
	}

	/*
	 * If we were on the last char of a line that got shifted left,
	 * then move left one so we aren't beyond the end of the line
	 */
#ifdef KANJI
	if (gchar(Curschar) == NUL && Curschar->index > 0)
		kdec(Curschar);
#else
	if (gchar(Curschar) == NUL && Curschar->index > 0)
		Curschar->index--;
#endif

	updatescreen();

	if (nlines > P(P_RP))
		smsg("%d lines %ced", nlines, opchar);
}

/*
 * dodelete - handle a delete operation
 */
int
dodelete(c1, c2, num)
#ifdef KANJI
unsigned int	c1, c2;
#else
uchar	c1, c2;
#endif
int	num;
{
	LPTR	top, bot;
	int	nlines, ret;
	register int	n;

	ret = doyank();
	/*
	 * Do a yank of whatever we're about to delete. If there's too much
	 * stuff to fit in the yank buffer, then get a confirmation before
	 * doing the delete. This is crude, but simple. And it avoids doing
	 * a delete of something we can't put back if we want.
	 */
	if (ret == EXCEED) {
		msg("yank buffer exceeded: press <y> to confirm");
		if (vgetc() != 'y') {
			emsg("delete aborted");
			*Curschar = startop;
			return EXCEED;
		}
	}
	if (ret == FAILED) {
		*Curschar = startop;
		return FAILED;
	}

	top = startop;
	bot = *Curschar;

	if (lt(&bot, &top))
		pswap(&top, &bot);

	u_save(top.linep->prev, bot.linep->next);

	nlines = cntllines(&top, &bot);
	*Curschar = top;
	cursupdate();

	if (mtype == MLINE) {
		delline(nlines, TRUE);
	} else {
		if (!mincl) {
#ifdef KANJI
			if (bot.index == 0)
				kdec(&bot);
			if (bot.index != 0)
				kdec(&bot);
#else
			if (bot.index == 0)
				dec(&bot);
			if (bot.index != 0)
				dec(&bot);
#endif
		}

		if (!mincl && top.linep == bot.linep && lineempty()) {
			delline(1, TRUE);
			Curschar->index = 0;
		} else if (top.linep == bot.linep) {	/* del. within line */
			n = bot.index - top.index + 1;
#ifdef KANJI
			while (n-- > 0) {
				if (isKanji(gchar(Curschar)))
					n--;
				if (!delchar(TRUE))
					break;
			}
#else
			while (n--)
				if (!delchar(TRUE))
					break;
#endif
		} else {				/* del. between lines */
			n = Curschar->index;
			while (Curschar->index >= n)
				if (!delchar(TRUE))
					break;

			top = *Curschar;
			*Curschar = *nextline(Curschar);
			delline(nlines-2, TRUE);
			Curschar->index = 0;
			n = bot.index + 1;
#ifdef KANJI
			while (n-- > 0) {
				if (isKanji(gchar(Curschar)))
					n--;
				if (!delchar(TRUE))
					break;
			}
#else
			while (n--)
				if (!delchar(TRUE))
					break;
#endif
			*Curschar = top;
			if (dojoin(FALSE) == FALSE) {
				u_undo();
				return FAILED;
			}
			if (Curschar->index > 0)
				oneright();
		}
	}

	/* construct Redo buff */
#ifdef KANJI
	if (num != 0)
		sprintf(Redobuff, "d%d%c%s", num, c1, mkstr(c2));
	else
		sprintf(Redobuff, "d%c%s", c1, mkstr(c2));
#else
	if (num != 0)
		sprintf(Redobuff, "d%d%c%c", num, c1, c2);
	else
		sprintf(Redobuff, "d%c%c", c1, c2);
#endif

	set_want_col = TRUE;
	if (mtype == MCHAR && nlines == 1)
		updateline();
	else
		updatescreen();

	if (nlines > P(P_RP))
		smsg("%d fewer lines", nlines);

	return DONE;
}

/*
 * dofilter - handle a filter operation
 */

#define	ITMP	"viXXXXXX"
#define	OTMP	"voXXXXXX"

static	uchar	itmp[MAXPLEN];
static	uchar	otmp[MAXPLEN];


/*
 * dofilter - filter lines through a command given by the user
 *
 * We use temp files and the system() routine here. This would normally
 * be done using pipes on a UNIX machine, but this is more portable to
 * the machines we usually run on. The system() routine needs to be able
 * to deal with redirection somehow, and should handle things like looking
 * at the PATH env. variable, and adding reasonable extensions to the
 * command name given by the user. All reasonable versions of system()
 * do this.
 */
void
dofilter(c1, c2, num)
#ifdef KANJI
unsigned int	c1, c2;
#else
uchar	c1, c2;
#endif
int	num;
{
	char	*mktemp();
	static	uchar	*lastcmd = NULL;/* the last thing we did */
	uchar	*buff;			/* cmd buffer from getcmdln() */
	uchar	cmdln[200];		/* filtering command line */
	LPTR	top, bot;
	int	nlines;
	uchar	*tmpdir, *getenv();

	top = startop;
	bot = *Curschar;

	buff = getcmdln('!');

	if (buff == NULL)	/* user backed out of the command prompt */
		return;

	if (*buff == '!') {		/* use the 'last' command */
		if (lastcmd == NULL) {
			emsg("No previous command");
			return;
		}
		buff = lastcmd;
	}

	/*
	 * Remember the current command
	 */
	if (lastcmd != NULL)
		free(lastcmd);
	lastcmd = strsave(buff);

	if (lt(&bot, &top))
		pswap(&top, &bot);

	u_save(top.linep->prev, bot.linep->next);

	nlines = cntllines(&top, &bot);
	*Curschar = top;
	cursupdate();

	/*
	 * 1. Form temp file names
	 * 2. Write the lines to a temp file
	 * 3. Run the filter command on the temp file
	 * 4. Read the output of the command into the buffer
	 * 5. Delete the original lines to be filtered
	 * 6. Remove the temp files
	 */

	tmpdir = getenv("TMP");
	strcpy(itmp, mkpath(tmpdir, ITMP));
	strcpy(otmp, mkpath(tmpdir, OTMP));

	if (mktemp(itmp) == NULL || mktemp(otmp) == NULL) {
		emsg("Can't get temp file names");
		return;
	}

	if (!writeit(itmp, &top, &bot, FALSE)) {
		emsg("Can't create input temp file");
		return;
	}

	sprintf(cmdln, "%s <%s >%s", buff, itmp, otmp);

	if (system(cmdln) != 0) {
		emsg("Filter command failed");
		/* remove(ITMP); bug in original 3.69 - Junn */
		remove(itmp);
		return;
	}

	if (readfile(otmp, &bot, TRUE)) {
		emsg("Can't read filter output");
		return;
	}

	delline(nlines, TRUE);

	remove(itmp);
	remove(otmp);

	/* construct Redo buff */
#ifdef KANJI
	if (num != 0)
		sprintf(Redobuff, "d%d%c%s", num, c1, mkstr(c2));
	else
		sprintf(Redobuff, "d%c%s", c1, mkstr(c2));
#else
	if (num != 0)
		sprintf(Redobuff, "d%d%c%c", num, c1, c2);
	else
		sprintf(Redobuff, "d%c%c", c1, c2);
#endif

	updatescreen();

	if (nlines > P(P_RP))
		smsg("%d lines filtered", nlines);
}

#ifdef	TILDEOP
void
dotilde(c1, c2, num)
#ifdef KANJI
unsigned int	c1, c2;
#else
uchar	c1, c2;
#endif
int	num;
{
	LPTR	top, bot;
	register uchar	c;

	/* construct Redo buff */
#ifdef KANJI
	if (num != 0)
		sprintf(Redobuff, "~%d%c%s", num, c1, mkstr(c2));
	else
		sprintf(Redobuff, "~%c%s", c1, mkstr(c2));
#else
	if (num != 0)
		sprintf(Redobuff, "~%d%c%c", num, c1, c2);
	else
		sprintf(Redobuff, "~%c%c", c1, c2);
#endif

	top = startop;
	bot = *Curschar;

	if (lt(&bot, &top))
		pswap(&top, &bot);

	u_save(top.linep->prev, bot.linep->next);

	if (mtype == MLINE) {
		top.index = 0;
		bot.index = strlen(bot.linep->s);
	} else {
#ifdef KANJI
		if (!mincl && bot.index != 0)
			kdec(&bot);
#else
		if (!mincl) {
			if (bot.index)
				bot.index--;
		}
#endif
	}

#ifdef KANJI
	for (; ltoreq(&top, &bot) ;kinc(&top)) {
		/*
		 * Swap case through the range
		 */
		c = gchar(&top);
		if (isascii(c) && isalpha(c)) {
			if (islower(c))
				c = toupper(c);
			else
				c = tolower(c);

			pchar(&top, c);		/* Change current character. */
			CHANGED;
		}
	}
#else
	for (; ltoreq(&top, &bot) ;inc(&top)) {
		/*
		 * Swap case through the range
		 */
		c = gchar(&top);
		if (isalpha(c)) {
			if (islower(c))
				c = toupper(c);
			else
				c = tolower(c);

			pchar(&top, c);		/* Change current character. */
			CHANGED;
		}
	}
#endif
	*Curschar = startop;
	updatescreen();
}
#endif /*TILDEOP*/

/*
 * dochange - handle a change operation
 */
void
dochange(c1, c2, num)
#ifdef KANJI
unsigned int	c1, c2;
#else
uchar	c1, c2;
#endif
int	num;
{
	uchar	sbuf[16];
	bool_t	doappend;	/* true if we should do append, not insert */
	bool_t	at_eof;		/* changing through the end of file */
	LPTR	top, bot;

	top = startop;
	bot = *Curschar;

	if (lt(&bot, &top))
		pswap(&top, &bot);

	doappend = mincl && endofline(&bot);
	at_eof = (bot.linep->next == Fileend->linep);

	if (dodelete(c1, c2, num) != DONE)
		return;

	if (mtype == MLINE) {
		/*
		 * If we made a change through the last line of the file,
		 * then the cursor got backed up, and we need to open a
		 * new line forward, otherwise we go backward.
		 */
		if (at_eof)
			opencmd(FORWARD, FALSE);
		else
			opencmd(BACKWARD, FALSE);
	} else {
		if (doappend && !lineempty())
#ifdef KANJI
			kinc(Curschar);
#else
			inc(Curschar);
#endif
	}

	if (num)
		sprintf(sbuf, "c%d%c%c", num, c1, c2);
	else
		sprintf(sbuf, "c%c%c", c1, c2);

	startinsert(sbuf, mtype == MLINE);
}

static	uchar	ybuf[YBSIZE];
static	int	ybtype = MBAD;

int
doyank()
{
	LPTR	top, bot, obot;
	uchar	*yptr = ybuf;
	uchar	*ybend = &ybuf[YBSIZE-1];
	int	nlines;
	int	ret = DONE;

	top = startop;
	bot = obot = *Curschar;

	if (lt(&bot, &top))
		pswap(&top, &bot);

	nlines = cntllines(&top, &bot);

#ifdef KANJI

	if (mtype == MLINE) {
		top.index = 0;
		bot.index = strlen(bot.linep->s);
		/*
		 * The following statement checks for the special case of
		 * yanking a blank line at the beginning of the file. If
		 * not handled right, we yank an extra char (a newline).
		 */
		if (kdec(&bot) == -1) {
			ybuf[0] = NUL;
			goto out;
		}
	} else {
		if (!mincl) {
			if (bot.index == 0)
				kdec(&bot);
			if (bot.index != 0)
				kdec(&bot);
			if (lt(&bot, &top)) {
				ret = FAILED;
				goto out;
			}
		} else if (equal(&top, &bot) && gchar(&top) == NUL) {
			ret = FAILED;
			goto out;
		}
	}

	for (; ltoreq(&top, &bot) ;inc(&top)) {
		*yptr = (gchar(&top) != NUL) ? gchar(&top) : NL;
		if (isKanji(*yptr)) {
			inc(&top);
			*++yptr = (gchar(&top) != NUL) ? gchar(&top) : NL;
		}
		if (++yptr + 1 >= ybend) {
			emsg("yank too big for buffer");
			ret = EXCEED;
			goto out;
		}
	}

#else

	if (mtype == MLINE) {
		top.index = 0;
		bot.index = strlen(bot.linep->s);
		/*
		 * The following statement checks for the special case of
		 * yanking a blank line at the beginning of the file. If
		 * not handled right, we yank an extra char (a newline).
		 */
		if (dec(&bot) == -1) {
			ybuf[0] = NUL;
			goto out;
		}
	} else {
		if (!mincl) {
			if (bot.index == 0)
				dec(&bot);
			if (bot.index != 0)
				dec(&bot);
			if (lt(&bot, &top)) {
				ret = FAILED;
				goto out;
			}
		} else if (equal(&top, &bot) && gchar(&top) == NUL) {
			ret = FAILED;
			goto out;
		}
	}

	for (; ltoreq(&top, &bot) ;inc(&top)) {
		*yptr = (gchar(&top) != NUL) ? gchar(&top) : NL;
		if (++yptr >= ybend) {
			emsg("yank too big for buffer");
			ret = EXCEED;
			goto out;
		}
	}

#endif

	*yptr = NUL;
	if (operator == YANK && nlines > P(P_RP))
		smsg("%d lines yanked", nlines);

out:
	if ((ybtype = (ret == DONE)? mtype: MBAD) == MBAD)
		beep();
	*Curschar = (operator == YANK)? startop: obot;
	return ret;
}

/*
 * doput(dir)
 *
 * Put the yank buffer at the current location, using the direction given
 * by 'dir'.
 */
void
doput(dir)
int	dir;
{
	uchar	*s;

	if (ybtype == MBAD) {
		beep();
		return;
	}

	if (ybtype == MLINE) {
		u_saveline();
		if (!inslines(Curschar->linep, dir, ybuf, TRUE))
			u_undo();
		else
			CHANGED;
	} else {
		/*
		 * If we did a character-oriented yank, and the buffer
		 * contains multiple lines, the situation is more complex.
		 * For the moment, we punt, and pretend the user did a
		 * line-oriented yank. This doesn't actually happen that
		 * often.
		 */
		if (strchr(ybuf, NL) != NULL) {
			u_saveline();
			/*
			 * If the buffer ends with NL,
			 * eliminate it while inserting lines.
			 */
			s = ybuf + strlen(ybuf) - 1;
			if (*s == NL)
				*s = NUL;
			if (!inslines(Curschar->linep, dir, ybuf, TRUE))
				u_undo();
			else
				CHANGED;
			*s = NL;
			return;
		} else {
			int	len;

			len = strlen(Curschar->linep->s) + strlen(ybuf) + 1;
			if (len > MAXLINE - 1) {
				emsg("Line too long");
				return;
			}
			if ((s = alloc((unsigned) len)) == NULL)
				return;
			u_saveline();
			strcpy(s, Curschar->linep->s);
#ifdef KANJI
			if (dir == FORWARD && *s != NUL)
				kinc(Curschar);
#else
			if (dir == FORWARD && *s != NUL)
				Curschar->index++;
#endif
			strcpy(s + Curschar->index, ybuf);
			strcat(s, &Curschar->linep->s[Curschar->index]);
			free(Curschar->linep->s);
			Curschar->linep->s = s;
			Curschar->linep->size = len;
#ifdef KANJI
			Curschar->index += strlen(ybuf);
			kdec(Curschar);
#else
			Curschar->index += strlen(ybuf) - 1;
#endif
			updateline();
			CHANGED;
		}
	}
}

bool_t
dojoin(join_cmd)
bool_t	join_cmd;		/* handling a real "join" command? */
{
	int	scol;		/* save cursor column */
	int	size;		/* size of the joined line */
	int	empty;		/* the joinning line is empty */
	int	dot;		/* the joined line ends with '.' */

	if (nextline(Curschar) == NULL)		/* on last line */
		return FALSE;

	if (!canincrease(size = strlen(Curschar->linep->next->s)))
		return FALSE;

	empty = (Curschar->linep->s[0] == NUL);

	while (oneright())			/* to end of line */
		;

	dot = (!empty && Curschar->linep->s[Curschar->index] == '.');

	strcat(Curschar->linep->s, Curschar->linep->next->s);

	/*
	 * Delete the following line. To do this we move the cursor
	 * there briefly, and then move it back. Don't back up if the
	 * delete made us the last line.
	 */
	Curschar->linep = Curschar->linep->next;
	scol = Curschar->index;

	if (nextline(Curschar) != NULL) {
		delline(1, TRUE);
		Curschar->linep = Curschar->linep->prev;
	} else
		delline(1, TRUE);

	Curschar->index = scol;

	if (join_cmd && !empty)
		oneright();	/* go to first char. of joined line */

	if (join_cmd && size != 0) {
		/*
		 * Delete leading white space on the joined line
		 * and insert a single space.
		 */
#ifdef KANJI
		while (gchar(Curschar) == ' ' || gchar(Curschar) == TAB ||
				gchar(Curschar) == KS1 &&
				Curschar->linep->s[Curschar->index+1] == KS2)
			delchar(TRUE);
		if (!empty && gchar(Curschar) != ')' &&
				!isKanji(gchar(Curschar))) {
			inschar(' ');
			if (dot) {
				inschar(' ');
				oneleft();
			}
			oneleft();
		}
#else
		while (gchar(Curschar) == ' ' || gchar(Curschar) == TAB)
			delchar(TRUE);
		if (!empty && gchar(Curschar) != ')') {
			inschar(' ');
			if (dot) {
				inschar(' ');
				oneleft();
			}
			oneleft();
		}
#endif
	}

	return TRUE;
}

void
startinsert(initstr, startln)
uchar	*initstr;
int	startln;	/* if set, insert point really at start of line */
{
	register uchar	*p, c;

	tstpoff();
	*Insstart = *Curschar;
	if (startln)
		Insstart->index = 0;
	Ninsert = 0;
	Insptr = Insbuff;
	for (p=initstr; (c=(*p++))!='\0'; )
		*Insptr++ = c;

	can_ab = TRUE;
	if (*initstr == 'R')
		State = REPLACE;
	else
		State = INSERT;

	if (P(P_MO))
		mmsg((State == INSERT) ? "Insert Mode" : "Replace Mode");

#ifdef FEPCTRL
	if (P(P_FC))
		fep_on();
#endif
}

/*
 * whiteinout(inout,num)
 *
 * If inout==0, add a white_space to the begining of the next num lines.
 * If inout==1, delete a white_space from the beginning of the next num lines.
 */
static void
whiteinout(inout, num)
int	inout;
int	num;
{
	int		ntodo = num;
	LPTR		*p;
	register int	white, c;

	beginline(FALSE);
	while (ntodo-- > 0) {
		beginline(FALSE);
		Curschar->index = 0;
		white = 0;
		while ( ((c = gchar(Curschar)) == TAB) || (c == ' ') ) {
			delchar(TRUE);
			if (c == TAB) {
				white += P(P_TS);
			}
			else {
				white += 1;
			}
		}
		if ( (c != NUL) || (white > 0) ) {
			if (inout == 0) {
				white += P(P_SW);
			}
			else {
				white -= P(P_SW);
			}
			while (white >= P(P_TS)) {
				inschar(TAB);
				white -= P(P_TS);
			}
			while (white > 0) {
				inschar(' ');
				white--;
			}
		}

		if ( ntodo > 0 ) {
			if ((p = nextline(Curschar)) != NULL)
				*Curschar = *p;
			else
				break;
		}
	}
}

/*
 * inslines(lp, dir, buf, upd)
 *
 * Inserts lines in the file from the given buffer. Lines are inserted
 * before or after "lp" according to the given direction flag. Newlines
 * in the buffer result in multiple lines being inserted. The cursor
 * is left on the first of the inserted lines, if "upd"ating screen.
 */
bool_t
inslines(lp, dir, buf, upd)
LINE	*lp;
int	dir;
uchar	*buf;
bool_t	upd;
{
	register uchar	*cp = buf;
	register int	len;
	uchar	*ep;
	LINE	*l, *nc = NULL;

	if (dir == BACKWARD)
		lp = lp->prev;

	do {
		if ((ep = (uchar *)strchr(cp, NL)) == NULL)
			len = strlen(cp);
		else
			len = ep - cp;

		l = newline(len);
		if (l == NULL)
			return FALSE;

		if (len != 0)
			strncpy(l->s, cp, len);
		l->s[len] = NUL;

		l->next = lp->next;
		l->prev = lp;
		lp->next->prev = l;
		lp->next = l;

		if (nc == NULL)
			nc = l;

		lp = lp->next;

		cp = ep + 1;
	} while (ep != NULL);

	if (dir == BACKWARD)	/* fix the top line in case we were there */
		Filemem->linep = Filetop->linep->next;

	renum();

	if (upd) {
		updatescreen();
		Curschar->linep = nc;
		Curschar->index = 0;
	}

	return TRUE;
}
