/* $Header: /nw/tony/src/stevie/src/RCS/os2.c,v 1.7 89/08/07 05:49:19 tony Exp $
 *
 * OS/2 System-dependent routines.
 */

#define	INCL_BASE
#include "stevie.h"
#include <os2.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

uchar	*Initfile = "stevie.rc";

/*
 * inchar() - get a character from the keyboard
 */
int
inchar()
{
	register int	c;

	got_int = FALSE;

	for (;;beep()) {	/* loop until we get a valid character */

		flushbuf();	/* flush any pending output */

#ifdef KANJI
		c = getch();
		if (isKanji(c)) {
			return (c << 8) | getch();
		} else {
			return c;
		}
#else
		switch (c = getch()) {
		case 0x1e:
			return K_CGRAVE;
		case 0:				/* special key */
			if (State != NORMAL) {
				c = getch();	/* throw away next char */
				continue;	/* and loop for another char */
			}
			switch (c = getch()) {
			case 0x50:
				return K_DARROW;
			case 0x48:
				return K_UARROW;
			case 0x4b:
				return K_LARROW;
			case 0x4d:
				return K_RARROW;
			case 0x52:
				return K_INSERT;
			case 0x47:
				stuffin("1G");
				return -1;
			case 0x4f:
				stuffin("G");
				return -1;
			case 0x51:
				stuffin(mkstr(Ctrl('F')));
				return -1;
			case 0x49:
				stuffin(mkstr(Ctrl('B')));
				return -1;
			/*
			 * Hard-code some useful function key macros.
			 */
			case 0x3b: /* F1 */
				stuffin(":N\n");
				return -1;
			case 0x54: /* SF1 */
				stuffin(":N!\n");
				return -1;
			case 0x3c: /* F2 */
				stuffin(":n\n");
				return -1;
			case 0x55: /* SF2 */
				stuffin(":n!\n");
				return -1;
			case 0x3d: /* F3 */
				return K_CGRAVE;
			case 0x3e: /* F4 */
				stuffin(":rew\n");
				return -1;
			case 0x57: /* SF4 */
				stuffin(":rew!\n");
				return -1;
			case 0x3f: /* F5 */
				stuffin("[[");
				return -1;
			case 0x40: /* F6 */
				stuffin("]]");
				return -1;
			case 0x41: /* F7 - explain C declaration */
				stuffin("yyp^iexplain \033!!cdecl\n");
				return -1;
			case 0x42: /* F8 - declare C variable */
				stuffin("yyp!!cdecl\n");
				return -1;
			case 0x43: /* F9 */
				stuffin(":x\n");
				return -1;
			case 0x44: /* F10 */
				stuffin(":help\n");
				return -1;
			default:
				break;
			}
			break;

		default:
			return c;
		}
#endif
	}
}

static	uchar	outbuf[OBSIZE];
static	int	bpos = 0;

void
flushbuf()
{
	if (bpos != 0)
		write(1, outbuf, bpos);
	bpos = 0;
}

/*
 * Macro to output a character. Used within this file for speed.
 */
#define	outone(c)	outbuf[bpos++] = c; if (bpos >= OBSIZE) flushbuf()

/*
 * Function version for use outside this file.
 */
void
outchar(c)
register uchar	c;
{
	outbuf[bpos++] = c;
	if (bpos >= OBSIZE)
		flushbuf();
}

static	uchar	cell[2] = { 0, 7 };

/*
 * outstr(s) - write a string to the console
 *
 * We implement insert/delete line escape sequences here. This is kind
 * of a kludge, but at least it's localized to a single point.
 */
void
outstr(s)
register char	*s;
{
#ifdef TERMCAP
	while (*s) {
		outone(*s++);
	}
#else
	if (*s != ESC) {
		while (*s) {
			outone(*s++);
		}
	}
	if (T_DL && strcmp(s, T_DL) == 0) {	/* delete line */
		int	r, c;

		flushbuf();
		VioGetCurPos(&r, &c, 0);
		VioScrollUp(r, 0, 100, 100, 1, cell, 0);
		return;
	}
	if (T_IL && strcmp(s, T_IL) == 0) {	/* insert line */
		int	r, c;

		flushbuf();
		VioGetCurPos(&r, &c, 0);
		VioScrollDn(r, 0, 100, 100, 1, cell, 0);
		return;
	}
	while (*s) {
		outone(*s++);
	}
#endif
}

void
beep()
{
	outone('\007');
}

sleep(n)
int	n;
{
	DosSleep(1000L * n);
}

void
delay()
{
	flushbuf();
	DosSleep(300L);
}

void
sig()
{
	signal(SIGINT, sig);

	got_int = TRUE;
}

void
tstpon()
{
	/* nothing to do here */
}

void
tstpoff()
{
	/* nothing to do here */
}

void
windinit()
{
#ifdef TERMCAP
	switch (t_init()) {
	case 1:
		break;
	case 0:
		fprintf(stderr, "insufficient terminal capability\n");
		exit(1);
	case -1:
		Columns = 80;
		P(P_LI) = Rows = 25;
		T_CM = "\033[%i%d;%dH";
		T_EB = "\033[J";
		T_EL = "\033[K";
		T_IL = "\033[L";
		T_DL = "\033[M";
		T_ED = "\033[2J";
		T_SC = "\033[s";
		T_RC = "\033[u";
		T_SO = "\033[7m";
		T_SE = "\033[m";
		T_US = "\033[4m";
		T_UE = "\033[m";
		T_CI = NULL;
		T_CV = NULL;
		T_TI = NULL;
		T_TE = NULL;
		break;
	}
#else
	Columns = 80;
	P(P_LI) = Rows = 25;
#endif

	signal(SIGINT, sig);
	if (T_TI) toutstr(T_TI, 1);
}

void
windexit(r)
int r;
{
	if (T_TE) toutstr(T_TE, 1);
	flushbuf();
	exit(r);
}

void
windgoto(r, c)
register int	r, c;
{
#ifdef TERMCAP
	char	*tgoto();

	toutstr(tgoto(T_CM, c, r), 1);
#else
	r++;
	c++;

	/*
	 * Check for overflow once, to save time.
	 */
	if (bpos + 8 >= OBSIZE)
		flushbuf();

	outbuf[bpos++] = '\033';
	outbuf[bpos++] = '[';
	if (r >= 10) {
		outbuf[bpos++] = r / 10 + '0';
		outbuf[bpos++] = r % 10 + '0';
	} else
		outbuf[bpos++] = r + '0';
	outbuf[bpos++] = ';';
	if (c >= 10) {
		outbuf[bpos++] = c / 10 + '0';
		outbuf[bpos++] = c % 10 + '0';
	} else
		outbuf[bpos++] = c + '0';
	outbuf[bpos++] = 'H';
#endif
}

FILE *
fopenb(fname, mode)
uchar	*fname;
uchar	*mode;
{
	FILE	*fopen();
	uchar	modestr[16];

	sprintf(modestr, "%sb", mode);
	return fopen(fname, modestr);
}

int
filestat(fname)
uchar	*fname;
{
	int	ret;
	struct stat stbuf;

	if (stat(fname, &stbuf) < 0)
		return 0;

	ret = FS_EXIST;
	if (stbuf.st_mode & S_IREAD)
		ret |= FS_READ;
	if (stbuf.st_mode & S_IWRITE)
		ret |= FS_WRITE;
	if (stbuf.st_mode & S_IFDIR)
		ret |= FS_DIR;

	return ret;
}

/*
 * Rename(of, nf) - rename existing file 'of' to 'nf'
 */
int
Rename(of, nf)
uchar	*of, *nf;
{
	unlink(nf);
	return rename(of, nf);
}

/*
 * backupname(fname, ext) - generate filename for backup
 */
uchar *
backupname(fname, ext)
uchar	*fname, *ext;
{
	register uchar	*s;
	static uchar	bname[MAXPLEN];

	strcpy(bname, fname);
	for (s = bname; *s; s++)
		/* void */;
	while (s > bname && *s != '/' && *s != '\\' && *s != ':')
		s--;
	while (*s && *s != '.')
		s++;
	*s = NUL;
	strcat(bname, ext);
	return bname;
}

/*
 * fixname(s) - fix up a dos name
 *
 * Takes a name like:
 *
 *	\x\y\z\base.ext
 *
 * and trims 'base' to 8 characters, and 'ext' to 3.
 */
uchar *
fixname(s)
uchar	*s;
{
	static	uchar	f[MAXPLEN];
	uchar	base[32];
	uchar	ext[32];
	uchar	*p;
	int	i;

	strcpy(f, s);

	for (i=0; i < MAXPLEN ;i++)
		if (f[i] == '/')
			f[i] = '\\';

	/*
	 * Split the name into directory, base, extension.
	 */
	if ((p = strrchr(f, '\\')) != NULL || (p = strrchr(f, ':')) != NULL) {
		strcpy(base, p+1);
		p[1] = '\0';
	} else {
		strcpy(base, f);
		f[0] = '\0';
	}

	if ((p = strchr(base, '.')) != NULL) {
		strcpy(ext, p+1);
		*p = '\0';
	} else
		ext[0] = '\0';

	/*
	 * Trim the base name if necessary.
	 */
	if (strlen(base) > 8)
		base[8] = '\0';
	
	if (strlen(ext) > 3)
		ext[3] = '\0';

	/*
	 * Paste it all back together
	 */
	strcat(f, base);
	strcat(f, ".");
	strcat(f, ext);

	return f;
}

uchar *
mkpath(dir, file)
uchar	*dir, *file;
{
	int	n;
	static uchar	buf[MAXPLEN];

	n = 0;
	if (dir) {
		strcpy(buf, dir);
		n = strlen(buf);
		if (buf[n-1] != '/' && buf[n-1] != '\\')
			buf[n++] = '\\';
	}
	strcpy(buf+n, file);
	return buf;
}

void
doshell(cmd)
uchar	*cmd;
{
	if (cmd == NULL)
		cmd = "cmd.exe";

	outstr("\r\n");
	flushbuf();
	system(cmd);
	wait_return();
}

void
cmdread(cmd)
uchar	*cmd;
{
	int fd, old, err;
	char tmpfile[MAXPLEN];

	if (cmd[0] == NUL) {
		emsg("Exec failed");
		return;
	}
	tmpnam(tmpfile);
	if ((fd = creat(tmpfile, 0600)) < 0) {
		emsg("Can't create temporary file");
		return;
	}
	old = dup(1);
	dup2(fd, 1);
	close(fd);
	outstr("\r");
	flushbuf();
	err = system(cmd);
	close(1);
	dup2(old, 1);
	if (err < 0)
		emsg("Exec failed");
	else {
		if (readfile(tmpfile, Curschar, 1))
			emsg("Can't open temporary file");
		else {
			updatescreen();
			CHANGED;
		}
	}
	remove(tmpfile);
}
