/*
 * pc98_50l.inc - nec pc9801 dependent
 *
 * Use graphics planes with 6x8 jis fonts and compressed 16x16 kanji fonts.
 *
 * This file should be included in dos.c when PC98_50LINE is defined in env.h
 *
 * Anyone can use this source file in any purpose.
 *
 * Oct. 30, 1989, koji ishii
 *
 * Nov. 6, 1989, koji ishii, optimized for msc5.1:
 *   Optimizings are done using -Fc assembly output option of msc5.1.
 *    o msc5.1 works better with arrays than with pointers in large model if
 *	its subscript is in register variable.
 *    o	msc5.1 generates poor codes for c = cp[i++]:  It will save i to ax,
 *	increment i, and then load to c.  In that case, as ax register
 *	cannot use as an index register, add instruction will be used.
 *	c = cp[i]; i++ will work better.
 */

#ifdef isKanji
#define	iskanji	isKanji
#else
#include <jctype.h> /* iskanji() */
#endif
#ifndef __TURBOC__
#include <memory.h> /* memcpy() */
#endif

static unsigned short near
ms2jis( register unsigned short c )
{
	register unsigned short c2;

	if (!(c & 0xff00))
		return 0;

	c2 = c & 0xff; c >>= 8;
	if (c <= 0x9f)
		c -= 0x71;
	else
		c -= 0xb1;
	c = (c << 1) | 1;
	if (c2 > 0x7f)
		c2--;
	if (c2 < 0x9e){
		c2 -= 0x1f;
	} else {
		c2 -= 0x7d;
		c++;
	}
	return (c << 8) | c2;
}

typedef struct {
	unsigned char bitmap[32];
	unsigned short code;
	unsigned char widthByte;
	unsigned char codelen;
} fontchar;

static void near
_getfontchar( register fontchar* pf )
{
	register int i;
	void far *pfar;
	union REGS reg;
	char buf[34];

	reg.h.ah = 0x14; pfar = (void far*)buf;
	reg.x.bx = FP_SEG( pfar ); reg.x.cx = FP_OFF( pfar );

	reg.x.dx = ms2jis( pf->code );
	if (reg.x.dx){
		pf->codelen = 2;
		int86( 0x18, &reg, &reg );
		for (i = 0; i < 8; i++)
			pf->bitmap[i  ] = buf[i*4+2] | buf[i*4+4];
		for (i = 0; i < 8; i++)
			pf->bitmap[i+8] = buf[i*4+3] | buf[i*4+5];
	} else {
		pf->codelen = 1;
		reg.x.dx = pf->code /* | 0x8000 */ ;
			/* | 0x8000 will take 7x13 fonts instead of 6x8 */
		int86( 0x18, &reg, &reg );
		memcpy( pf->bitmap, buf + 2, 8 );
	}
	pf->widthByte = buf[1];
}

static fontchar font1[256];

#define	getfontchar1( c )	(&(font1[c]))

static void near
preloadfont()
{
	register int i;
	extern char *getenv(const char *);
	char* homep = getenv( "HOME" );
	char* tmpp = getenv( "TMP" );
	FILE* fp = NULL;
	char writefont = 0;
	char loadfont = 1;
	char fn[65];
	char* fontfile = "%s/stevie.fon";
	char* readmode = "rb";

	if (tmpp){
		sprintf( fn, fontfile, tmpp );
		fp = fopen( fn, readmode );
	}
	if ((fp == NULL)&&(homep)){
		sprintf( fn, fontfile, homep );
		fp = fopen( fn, readmode );
	}
	if (fp == NULL){
		sprintf( fn, fontfile, "/etc" );
		fp = fopen( fn, readmode );
		if ((fp == NULL)&&(tmpp)){
			sprintf( fn, fontfile, tmpp );
			writefont = 1;
		}
	}
	if (fp != NULL){
		loadfont = 0;
		for (i = 0; i < 256; i++){
			if (fread( font1[i].bitmap, 1, 8, fp ) != 8){
				writefont = loadfont = 1;
				break;
			}
		}
		fclose( fp );
	}
	if (loadfont){
		printf( "wait, now loading font..." );
		for (i = 0; i < 256; i++){
			font1[i].code = i;
			_getfontchar( &font1[i] );
		}
		for (i = 0; i < 8; i++) /* create a back slash font */
			font1[0x5c].bitmap[i] = font1[0x2f].bitmap[7-i];
		printf( "ok\n" );
	}
	if (writefont){
		fp = fopen( fn, "wb" );
		for (i = 0; i < 256; i++)
			if (fwrite( font1[i].bitmap, 1, 8, fp ) != 8)
				break;
		if (ferror( fp )){
			fclose( fp );
			unlink( fn );
		} else
			fclose( fp );
	}
}

#define	nFontCache	97
static fontchar font2[nFontCache];

static fontchar* near
getfontchar2( register unsigned short c )
{
	fontchar* pf;

	pf = font2 + ((((c >> 8) & 0xff) + (c & 0xff)) % nFontCache);
	if (pf->code != c){
		pf->code = c;
		_getfontchar( pf );
	}
	return pf;
}

#define	fontheight	8

#define	PFRAME	((char far*)0xa8000000L)

static struct cursor_struct {
	char far* p;
	int x, y;
	unsigned standout: 1;
	unsigned should: 1;
	unsigned is: 1;
} cursor;

#define	linelen		(80 * fontheight)

#define	ADJCURSOR()	(cursor.p = (cursor.y * linelen) + cursor.x + PFRAME)

static void near
drawcursor()
{
	register int i;

	for (i = (fontheight-1)*80; i >= 0; i -= 80)
		cursor.p[i] = ~cursor.p[i];
	cursor.is ^= 1;
}

static void near
cls()
{
	register int i;

	for (i = 0; i < 32000; i++)
		PFRAME[i] = 0;
	cursor.x = cursor.y = 0; ADJCURSOR();
	cursor.is = 0;
}

#ifdef FEPCTRL
static void near
cls_text()
{
	unsigned char* s = "\033[2J";
	while (*s)
		bdos(6, *s++, 0);
}
#endif

static unsigned char lastmode = 0xff;

static void near
code_access_kcg()
{
	union REGS reg;

	if (lastmode == 0xff)
		return;

	reg.h.ah = 0x0a; /* set text crt mode */
	reg.h.al = lastmode;
	int86( 0x18, &reg, &reg );
}

static void near
dot_access_kcg()
{
	union REGS reg;

	if (lastmode == 0xff)
		return;

	reg.h.ah = 0x0a; /* set text crt mode */
	reg.h.al = lastmode | 0x08; /* dot access to kanji-cg */
	int86( 0x18, &reg, &reg );
}

static void near
startscreen()
{
	union REGS reg;

	reg.h.ah = 0x0b; /* sense text crt mode */
	int86( 0x18, &reg, &reg );
	lastmode = reg.h.al & 0x0f;

#ifdef FEPCTRL
	cls_text(); /* clear text crt */
#else
	reg.h.ah = 0x0d; /* stop text crt */
	int86( 0x18, &reg, &reg );
#endif

	dot_access_kcg();

	reg.h.ah = 0x12; /* stop text-cursor */
	int86( 0x18, &reg, &reg );

	reg.h.ah = 0x42; reg.h.ch = 0xc0; /* set graph to 640x400 mode */
	int86( 0x18, &reg, &reg );
	outp( 0x6a, 0 ); /* switch to 8 colors mode */
	outp( 0xa8, 0x77 ); outp( 0xaa, 0x77 ); /* setup color-palette */
	outp( 0xac, 0x00 ); outp( 0xae, 0x00 ); /* setup color-palette */

	reg.h.ah = 0x40; /* start graph crt */
	int86( 0x18, &reg, &reg );
}

static void near
stopscreen()
{
	union REGS reg;

	reg.h.ah = 0x41; /* stop graph crt */
	int86( 0x18, &reg, &reg );

	code_access_kcg();
	lastmode = 0xff;

#ifndef FEPCTRL
	reg.h.ah = 0x0c; /* start text crt */
	int86( 0x18, &reg, &reg );
#endif

	reg.h.ah = 0x11; /* start text-cursor */
	int86( 0x18, &reg, &reg );
}

static void near
openscreen()
{
	preloadfont();

	memset( PFRAME, '\0', 32000 );

	startscreen();

	cursor.should = 1; cls();

	{ fontchar dummy; dummy.code = 0x8140; _getfontchar( &dummy ); }
}

#define	closescreen()	stopscreen()

static char* near
doesc( unsigned char* cp )
{
	static struct cursor_struct lastsave;
	register int i, j;
	extern void rmemcpy( char*, char*, int );

	j = 1; /* means: j = 0; switch( cp[j++] )... */
	switch( cp[0] ){
	case '[':
		for (i = 0; isdigit( cp[j] ); j++)
			i = i * 10 + cp[j] - '0';
		cp += j;
		switch( *cp++ ){
		case 'K': /* i should be 0 here */
			if (cursor.x){
				j = 80 - cursor.x;
				for (i = (fontheight-1)*80; i >= 0; i -= 80)
					memset( cursor.p+i, '\0', j );
			} else
				memset( cursor.p, '\0', linelen );
			break;
		case 'J': /* i should be 2 here */
			cls();
			break;
		case 's': lastsave = cursor; break;
		case 'u': cursor = lastsave; break;
		case 'm': /* i should be 0 or 7 */
			cursor.standout = (i == 7);
			break;
		case 'L': cursor.x = 0; ADJCURSOR();
			if (i == 0)
				i = 1;
			j = (50 - cursor.y - i) * linelen;
			rmemcpy( cursor.p + linelen * i, cursor.p, j );
			memset( cursor.p, '\0', linelen * i );
			break;
		case 'M': cursor.x = 0; ADJCURSOR();
			if (i == 0)
				i = 1;
			j = (50 - cursor.y - i) * linelen;
			memcpy( cursor.p, cursor.p + linelen * i, j );
			memset( cursor.p + j, '\0', linelen * i );
			break;
		default: return cp - 1;
		}
		break;
	}
	return cp;
}

static void near
donewline()
{
	if (cursor.y == 49){
		cursor.y = 0; doesc( "[M" ); cursor.y = 48;
	}
	cursor.x = 0; cursor.y++; ADJCURSOR();
}

static void near
prnstr( unsigned char* cp )
{
	fontchar* pfc;
	register int i, c;
	extern void textblit( char far*, char*, int, int );

	if (cursor.is)
		drawcursor();
	i = 1; c = cp[0]; /* means: i = 0; c = cp[i++]; */
	for (;;){
		/* care that this loop assumes that iskanji( 0 ) == 0 */
		while (!iskanji( c )){
			switch( c ){
			case '\0': return;
			case '\a': putchar( '\a' ); break;
			case '\r': break;
			case '\n': donewline(); break;
			case '\033': cp = doesc( cp + i ); i = 0; break;
			case '\t':
				cursor.x = (cursor.x + 8) & ~7;
				if (cursor.x >= 80)
					donewline();
				ADJCURSOR();
				break;
			default:
				if (cursor.x >= 80)
					donewline();
				else
					cursor.x++;
				pfc = getfontchar1( c );
				textblit( cursor.p, pfc->bitmap,
				  cursor.standout, fontheight );
				cursor.p++;
				break;
			}
			c = cp[i]; i++;
		}
		while (iskanji( c )){
			pfc = getfontchar2( (c << 8) | cp[i] );
			cursor.x += pfc->widthByte;
			if (cursor.x > 80)
				donewline();
			textblit( cursor.p, pfc->bitmap,
			  cursor.standout, fontheight );
			cursor.p++;
			if (pfc->widthByte == 2){
				textblit( cursor.p, pfc->bitmap+fontheight,
				  cursor.standout, fontheight );
				cursor.p++;
			}
			c = cp[++i]; i++;
		}
	}
}
