/* $Header: /nw/tony/src/stevie/src/RCS/screen.c,v 1.8 89/08/02 09:26:33 tony Exp $
 *
 * Routines to manipulate the screen representations.
 */

#include "stevie.h"

/*
 * This gets set if we ignored an update request while input was pending.
 * We check this when the input is drained to see if the screen should be
 * updated.
 */
bool_t	need_redraw = FALSE;

/*
 * The following variable is set (in filetonext) to the number of physical
 * lines taken by the line the cursor is on. We use this to avoid extra
 * calls to plines(). The optimized routines lfiletonext() and lnexttoscreen()
 * make sure that the size of the cursor line hasn't changed. If so, lines
 * below the cursor will move up or down and we need to call the routines
 * filetonext() and nexttoscreen() to examine the entire screen.
 */
static	int	Cline_size;	/* size (in rows) of the cursor line */
static	int	Cline_row;	/* starting row of the cursor line */

static	uchar	*mkline();	/* calculate line string for "number" mode */

#ifndef PC98_VRAM

/*
 * filetonext()
 *
 * Based on the current value of Topchar, transfer a screenfull of
 * stuff from Filemem to Nextscreen, and update Botchar.
 */

static void
filetonext()
{
	register int	row, col;
	register uchar	*sp;
	LPTR	memp;
	LPTR	save;			/* save pos. in case line won't fit */
	uchar	extra[32];
	int	nextra = 0;
	register int	c = NUL;
	int	n;
	bool_t	done;		/* if TRUE, we hit the end of the file */
	bool_t	didline;	/* if TRUE, we finished the last line */
	int	srow;		/* starting row of the current line */
	int	lno;		/* number of the line we're doing */
	int	coff;		/* column offset */
#ifdef KANJI
	int	kanji2 = TRUE;	/* c has the 2nd byte of kanji */
#endif

	coff = P(P_NU) ? 8 : 0;

	save = memp = *Topchar;

	if (P(P_NU))
		lno = cntllines(Filemem, Topchar);

	done = FALSE;
	srow = row = col = 0;
	sp = Nextscreen[row];

	/*
	 * We go one past the end of the screen so we can find out if the
	 * last line fit on the screen or not.
	 */
	while (!done) {
#ifdef KANJI
		if (kanji2)
			kanji2 = FALSE;
		else if (isKanji(c))
			kanji2 = TRUE;
#endif

		if (col == 0 && P(P_NU) && memp.index == 0) {
			strcpy(extra, mkline(lno++));
			nextra = 8;
		}

		/* Get the next character to put on the screen. */

		/* The 'extra' array contains the extra stuff that is */
		/* inserted to represent special characters (tabs, and */
		/* other non-printable stuff.  The order in the 'extra' */
		/* array is reversed. */

		if ( nextra > 0 )
			c = extra[--nextra];
		else {
			/*
			 * c = (unsigned)(0xff & gchar(&memp));
			 * if (inc(&memp) == -1)
			 *	done = 1;
			 */
			c = memp.linep->s[memp.index++];
			if (c == NUL) {
				memp.linep = memp.linep->next;
				memp.index = 0;
				if (memp.linep == Fileend->linep)
					done = 1;
			}
			/* when getting a character from the file, we */
			/* may have to turn it into something else on */
			/* the way to putting it into 'Nextscreen'. */
			if ( c == TAB && !P(P_LS) ) {
				/* tab amount depends on current column */
				nextra = ((P(P_TS)-1) - (col - coff)%P(P_TS));
				for (n = 0; n < nextra; n++)
					extra[n] = ' ';
				c = ' ';
			}
			else if ( c == NUL && P(P_LS) ) {
				extra[0] = NUL;
				nextra = 1;
				c = '$';
			} else if ( (n = chars[c].ch_size) > 1 ) {
				uchar *p;
				nextra = 0;
				p = (uchar *)chars[c].ch_str;
				/* copy 'ch-str'ing into 'extra' in reverse */
				while ( n > 1 )
					extra[nextra++] = p[--n];
				c = p[0];
			}
		}

#ifdef KANJI
		if (c == NUL || col >= Columns  ||
			(col == Columns - 1 && kanji2 == FALSE && isKanji(c)))
#else
		if ( c == NUL || col >= Columns )
#endif
		{
			*sp = NUL;
			row++;
			col = 0;
			sp = Nextscreen[row];
			if (row >= Rows - 1) {
				didline = (c == NUL);
				break;
			}
			if (c == NUL) {
				/*
				 * Save this position in case the next
				 * line won't fit on the screen completely.
				 */
				srow = row;
				save = memp;
				continue;
			}
		}
		/* store the character in Nextscreen */
		*sp++ = c;
		col++;
	}
	/*
	 * If we didn't hit the end of the file, and we didn't finish
	 * the last line we were working on, then the line didn't fit.
	 */
	if (!done && !didline) {
		/*
		 * Clear the rest of the screen and mark the unused lines.
		 */
		for (; srow < Rows - 1;srow++) {
			Nextscreen[srow][0] = '@';
			Nextscreen[srow][1] = NUL;
		}
		*Botchar = save;
		return;
	}
	/* nul terminate the last line */
	if (done)
		*sp = NUL;
	/* make sure the rest of the screen is blank */
	/* put '~'s on rows that aren't part of the file. */
	if (col != 0)
		row++;
	while (row < Rows - 1) {
		Nextscreen[row][0] = '~';
		Nextscreen[row][1] = NUL;
		row++;
	}
	if (done)	/* we hit the end of the file */
		*Botchar = *Fileend;
	else
		*Botchar = memp;	/* FIX - prev? */
}

/*
 * lfiletonext() - like filetonext() but only for cursor line
 *
 * Returns true if the size of the cursor line (in rows) hasn't changed.
 * This determines whether or not we need to call filetonext() to examine
 * the entire screen for changes.
 */
static bool_t
lfiletonext()
{
	register int	row, col;
	register uchar	*sp;
	LPTR	memp;
	uchar	extra[32];
	int	nextra = 0;
	register int	c = NUL;
	int	n;
	bool_t	eof;
	int	lno;		/* number of the line we're doing */
	int	coff;		/* column offset */
#ifdef KANJI
	int	kanji2 = TRUE;	/* c has the 2nd byte of kanji */
#endif

	coff = P(P_NU) ? 8 : 0;

	/*
	 * This should be done more efficiently.
	 */
	if (P(P_NU))
		lno = cntllines(Filemem, Curschar);

	memp = *Curschar;
	memp.index = 0;

	eof = FALSE;
	col = 0;
	row = Cline_row;
	sp = Nextscreen[row];

	while (!eof) {
#ifdef KANJI
		if (kanji2)
			kanji2 = FALSE;
		else if (isKanji(c))
			kanji2 = TRUE;
#endif

		if (col == 0 && P(P_NU) && memp.index == 0) {
			strcpy(extra, mkline(lno));
			nextra = 8;
		}

		/* Get the next character to put on the screen. */

		/* The 'extra' array contains the extra stuff that is */
		/* inserted to represent special characters (tabs, and */
		/* other non-printable stuff.  The order in the 'extra' */
		/* array is reversed. */

		if ( nextra > 0 )
			c = extra[--nextra];
		else {
			/*
			 * c = (unsigned)(0xff & gchar(&memp));
			 * if (inc(&memp) == -1)
			 *	eof = TRUE;
			 */
			c = memp.linep->s[memp.index++];
			if (c == NUL) {
				memp.linep = memp.linep->next;
				memp.index = 0;
				if (memp.linep == Fileend->linep)
					eof = TRUE;
			}
			/* when getting a character from the file, we */
			/* may have to turn it into something else on */
			/* the way to putting it into 'Nextscreen'. */
			if ( c == TAB && !P(P_LS) ) {
				/* tab amount depends on current column */
				nextra = ((P(P_TS)-1) - (col - coff)%P(P_TS));
				for (n = 0; n < nextra; n++)
					extra[n] = ' ';
				c = ' ';
			} else if ( c == NUL && P(P_LS) ) {
				extra[0] = NUL;
				nextra = 1;
				c = '$';
			} else if ( c != NUL && (n=chars[c].ch_size) > 1 ) {
				uchar *p;
				nextra = 0;
				p = (uchar *)chars[c].ch_str;
				/* copy 'ch-str'ing into 'extra' in reverse */
				while ( n > 1 )
					extra[nextra++] = p[--n];
				c = p[0];
			}
		}

#ifdef KANJI
		if (c == NUL || col >= Columns  ||
			(col == Columns - 1 && kanji2 == FALSE && isKanji(c)))
#else
		if (c == NUL || col >= Columns)
#endif
		{
			/* blank out the rest of this row */
			*sp = NUL;
			row++;
			col = 0;
			/* get pointer to start of next row */
			sp = Nextscreen[row];
			if (c == NUL)
				break;
			if (row >= Rows - 1) {
				row++;
				break;
			}
		}
		/* store the character in Nextscreen */
		*sp++ = c;
		col++;
	}
	/* nul terminate the line */
	if (eof)
		*sp = NUL;
	return ((row - Cline_row) == Cline_size);
}

/*
 * nexttoscreen
 *
 * Transfer the contents of Nextscreen to the screen, using Realscreen
 * to avoid unnecessary output.
 */
static void
nexttoscreen()
{
	donexttoscreen(0, Rows - 1);
}

/*
 * lnexttoscreen
 *
 * Like nexttoscreen() but only for the cursor line.
 */
static void
lnexttoscreen()
{
	donexttoscreen(Cline_row, Cline_row + Cline_size);
}

/*
 * donexttoscreen
 */
donexttoscreen(from, upto)
int	from, upto;
{
	register uchar	*np, *rp;
	register int	row, col;
	int	gorow, gocol;

	if (anyinput()) {
		need_redraw = TRUE;
		return;
	}

	if (T_CI) toutstr(T_CI, 1);	/* disable cursor */

	gorow = gocol = -1;
	for (row = from; row < upto; row++) {
		col = 0;
#ifdef KANJI
		/*
		 * clear out kanji characters on Realscreen
		 * to force redraw
		 */
		rp = Realscreen[row];
		np = Nextscreen[row];
		while (*rp && *np) {
			if (isKanji(*rp)) {
				if (isKanji(*np)) {
					if (*rp != *np || rp[1] != np[1]) {
						*rp = rp[1] = BADCH;
					}
					rp++;
					np++;
				} else {
					*rp = rp[1] = BADCH;
				}
			} else if (isKanji(*np)) {
				if (isKanji(rp[1])) {
					rp[1] = rp[2] = BADCH;
				} if (rp[1] != NUL) {
					rp++;
					np++;
				}
			}
			rp++;
			np++;
		}
#endif
		rp = Realscreen[row];
		np = Nextscreen[row];
		while (*rp && *np) {
			/* If desired screen (contents of Nextscreen) does */
			/* not match what's really there, put it there. */
			if (*rp != *np) {
				/* if we are positioned at the right place, */
				/* we don't have to use windgoto(). */
				if (gorow != row || gocol + 4 < col)
					windgoto(gorow = row, gocol = col);
				else {
					/*
					 * If we're near to right col,
					 * don't send an entire esc. seq.
					 * (this happens a lot!)
					 */
					int	i = gocol - col;
					while (i < 0)
						outchar(np[i++]);
					gocol = col;
				}
				outchar(*rp = *np);
				gocol++;
			}
			rp++;
			np++;
			col++;
		}
		if (*np) {
			if (gocol != col || gorow != row)
				windgoto(gorow = row, gocol = col);
			while (*np) {
				outchar(*rp++ = *np++);
				col++;
			}
			*rp = NUL;
			if (col < Columns)
				toutstr(T_EL, 1);
		} else if (*rp) {
			*rp = NUL;
			if (gocol != col || gorow != row)
				windgoto(gorow = row, gocol = col);
			if (col < Columns)
				toutstr(T_EL, 1);
		}
	}

	if (T_CV) toutstr(T_CV, 1);	/* enable cursor again */
}

#else /*PC98_VRAM*/

uchar	stoj1[256] = {
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  1,  3,  5,  7,  9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29,
	 31, 33, 35, 37, 39, 41, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	 63, 65, 67, 69, 71, 73, 75, 77, 79, 81, 83, 85, 87, 89, 91, 93,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
};

uchar	stoj2[256] = {
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48,
	 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64,
	 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80,
	 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95,  0,
	 96, 97, 98, 99,100,101,102,103,104,105,106,107,108,109,110,111,
	112,113,114,115,116,117,118,119,120,121,122,123,124,125,126, 33,
	 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49,
	 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65,
	 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81,
	 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97,
	 98, 99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,
	114,115,116,117,118,119,120,121,122,123,124,125,126,  0,  0,  0
};

#define	A_NORM	0xe1	/* normal white; see dos.c */

#define	MK_FAR(s,o) (unsigned char far *)((unsigned long)(s)<<16|(unsigned)(o))
static	unsigned short far *vram  = (unsigned short far *)MK_FAR(0xa000, 0);
static	unsigned short far *vattr = (unsigned short far *)MK_FAR(0xa200, 0);

/*
 * filetonext()
 *
 * Based on the current value of Topchar, transfer a screenfull of
 * stuff from Filemem to PC-9801 Video RAM, and update Botchar.
 */

static void
filetonext()
{
	register int	c, pos;
	uchar	k1 = 0;
	int	savpos, endpos;
	int	col, n;
	LPTR	memp, save;
	uchar	*extra, exbuf[32];
	int	nextra = 0;
	bool_t	done;		/* if TRUE, we hit the end of the file */
	bool_t	didline;	/* if TRUE, we finished the last line */
	int	lno;		/* number of the line we're doing */
	int	coff;		/* column offset */

	coff = P(P_NU) ? 8 : 0;

	save = memp = *Topchar;

	if (P(P_NU))
		lno = cntllines(Filemem, Topchar);

	done = FALSE;
	col = pos = 0;
	endpos = (Rows - 1) * Columns;

	/*
	 * We go one past the end of the screen so we can find out if the
	 * last line fit on the screen or not.
	 */
	while (!done) {

		if (col == 0 && P(P_NU) && memp.index == 0) {
			strcpy(extra = exbuf, mkline(lno++));
			nextra = 8;
		}

		/* Get the next character to put on the screen. */

		/* The 'extra' array contains the extra stuff that is */
		/* inserted to represent special characters (tabs, and */
		/* other non-printable stuff.  The order in the 'extra' */
		/* array is reversed. */

		if (nextra > 0)
			c = extra[--nextra];
		else {
			c = memp.linep->s[memp.index++];
			if (c == NUL) {
				memp.linep = memp.linep->next;
				memp.index = 0;
				if (memp.linep == Fileend->linep)
					done = 1;
				if (P(P_LS)) {
					extra = "";
					nextra = 1;
					c = '$';
				}
			} else if (c == TAB && !P(P_LS)) {
				extra = "                               ";
				nextra = P(P_TS) - (col - coff) % P(P_TS) - 1;
				c = ' ';
			} else if (chars[c].ch_size > 1) {
				uchar *p;
				n = chars[c].ch_size;
				nextra = 0;
				extra = exbuf;
				p = (uchar *)chars[c].ch_str;
				/* copy 'ch-str'ing into 'extra' in reverse */
				while (n > 1)
					extra[nextra++] = p[--n];
				c = p[0];
			}
		}

		if (c == NUL || col >= Columns  ||
				(col == Columns-1 && k1 == 0 && isKanji(c))) {
			n = Columns - col;
			while (n-- > 0)
				vram[pos++] = ' ';
			col = 0;
			if (pos >= endpos) {
				didline = (c == NUL);
				break;
			}
			if (c == NUL) {
				/*
				 * Save this position in case the next
				 * line won't fit on the screen completely.
				 */
				savpos = pos;
				save = memp;
				continue;
			}
		}
		/* output the character to Video RAM */
		if (k1) {
			vram[pos++] = c = (stoj2[c]<<8)+stoj1[k1]+(c>=0x9f);
			vram[pos++] = c | 0x8000;
			k1 = 0;
		} else if (isKanji(c)) {
			k1 = c;
		} else {
			vram[pos++] = c;
		}
		col++;
	}
	/*
	 * If we didn't hit the end of the file, and we didn't finish
	 * the last line we were working on, then the line didn't fit.
	 */
	if (!done && !didline) {
		/*
		 * Clear the rest of the screen and mark the unused lines.
		 */
		pos = savpos;
		while (pos < endpos) {
			vram[pos++] = '@';
			n = Columns - 1;
			while (n--)
				vram[pos++] = ' ';
		}
		*Botchar = save;
		return;
	}
	/* nul terminate the last line */
	if (done && col > 0) {
		n = Columns - col;
		while (n--)
			vram[pos++] = ' ';
	}
	/* make sure the rest of the screen is blank */
	/* put '~'s on rows that aren't part of the file. */
	while (pos < endpos) {
		vram[pos++] = '~';
		n = Columns - 1;
		while (n--)
			vram[pos++] = ' ';
	}
	if (done)	/* we hit the end of the file */
		*Botchar = *Fileend;
	else
		*Botchar = memp;	/* FIX - prev? */
}

/*
 * lfiletonext() - like filetonext() but only for cursor line
 *
 * Returns true if the size of the cursor line (in rows) hasn't changed.
 * This determines whether or not we need to call filetonext() to examine
 * the entire screen for changes.
 */
static bool_t
lfiletonext()
{
	register int	pos, c;
	uchar	k1 = 0;
	int	startpos, endpos;
	int	col, n;
	LPTR	memp;
	uchar	*extra, exbuf[32];
	int	nextra = 0;
	bool_t	eof;
	int	lno;		/* number of the line we're doing */
	int	coff;		/* column offset */

	coff = P(P_NU) ? 8 : 0;

	/*
	 * This should be done more efficiently.
	 */
	if (P(P_NU))
		lno = cntllines(Filemem, Curschar);

	memp = *Curschar;
	memp.index = 0;

	eof = FALSE;
	col = 0;
	startpos = pos = Cline_row * Columns;
	endpos = (Rows - 1) * Columns;

	while (!eof) {

		if (col == 0 && P(P_NU) && memp.index == 0) {
			strcpy(extra = exbuf, mkline(lno));
			nextra = 8;
		}

		/* Get the next character to put on the screen. */

		/* The 'extra' array contains the extra stuff that is */
		/* inserted to represent special characters (tabs, and */
		/* other non-printable stuff.  The order in the 'extra' */
		/* array is reversed. */

		if (nextra > 0)
			c = extra[--nextra];
		else {
			c = memp.linep->s[memp.index++];
			if (c == NUL) {
				memp.linep = memp.linep->next;
				memp.index = 0;
				if (memp.linep == Fileend->linep)
					eof = TRUE;
				if (P(P_LS)) {
					extra = "";
					nextra = 1;
					c = '$';
				}
			} else if (c == TAB && !P(P_LS)) {
				/* tab amount depends on current column */
				extra = "                               ";
				nextra = P(P_TS) - (col - coff) % P(P_TS) - 1;
				c = ' ';
			} else if (chars[c].ch_size > 1) {
				uchar *p;
				n = chars[c].ch_size;
				nextra = 0;
				extra = exbuf;
				p = (uchar *)chars[c].ch_str;
				/* copy 'ch-str'ing into 'extra' in reverse */
				while (n > 1)
					extra[nextra++] = p[--n];
				c = p[0];
			}
		}

		if (c == NUL || col >= Columns  ||
				(col == Columns-1 && k1 == 0 && isKanji(c))) {
			/* blank out the rest of this row */
			n = Columns - col;
			while (n--)
				vram[pos++] = ' ';
			col = 0;
			/* get pointer to start of next row */
			if (c == NUL || pos >= endpos)
				break;
		}
		/* output the character to Video RAM */
		if (k1) {
			vram[pos++] = c = (stoj2[c]<<8)+stoj1[k1]+(c>=0x9f);
			vram[pos++] = c | 0x8000;
			k1 = 0;
		} else if (isKanji(c)) {
			k1 = c;
		} else {
			vram[pos++] = c;
		}
		col++;
	}
	/* nul terminate the line */
	if (eof && col > 0) {
		n = Columns - col;
		while (n--)
			vram[pos++] = ' ';
	}
	return (pos - startpos == Cline_size * Columns);
}

/*
 * nexttoscreen
 *
 * Transfer the contents of Nextscreen to the screen, using Realscreen
 * to avoid unnecessary output.
 */
static void
nexttoscreen()
{
	if (anyinput())
		need_redraw = TRUE;
	return;
}

/*
 * lnexttoscreen
 *
 * Like nexttoscreen() but only for the cursor line.
 */
static void
lnexttoscreen()
{
	if (anyinput())
		need_redraw = TRUE;
	return;
}

#endif /*PC98_VRAM*/

static uchar *
mkline(n)
register int	n;
{
	static	uchar	lbuf[9];
	register int	i = 2;

	strcpy(lbuf, "        ");
	do {
		lbuf[i++] = (n % 10) + '0';
		n /= 10;
	} while (n != 0);
	return lbuf;
}

/*
 * updateline() - update the line the cursor is on
 *
 * Updateline() is called after changes that only affect the line that
 * the cursor is on. This improves performance tremendously for normal
 * insert mode operation. The only thing we have to watch for is when
 * the cursor line grows or shrinks around a row boundary. This means
 * we have to repaint other parts of the screen appropriately. If
 * lfiletonext() returns FALSE, the size of the cursor line (in rows)
 * has changed and we have to call updatescreen() to do a complete job.
 */
void
updateline()
{
	if (!lfiletonext())
		updatescreen();	/* bag it, do the whole screen */
	else
		lnexttoscreen();
}

void
updatescreen()
{
	extern	bool_t	interactive;

	if (interactive) {
		filetonext();
		nexttoscreen();
	}
}

/*
 * prt_line() - print the given line
 */
void
prt_line(s)
uchar	*s;
{
	register int	si = 0;
	register int	c = NUL;
	register int	col = 0;

	uchar	extra[32];
	int	nextra = 0;
	int	n;
#ifdef KANJI
	int	kanji2 = TRUE;	/* c has the 2nd byte of kanji */
#endif

	for (;;) {
#ifdef KANJI
		if (kanji2)
			kanji2 = FALSE;
		else if (isKanji(c))
			kanji2 = TRUE;
#endif

		if ( nextra > 0 )
			c = extra[--nextra];
		else {
			c = s[si++];
			if ( c == TAB && !P(P_LS) ) {
				/* tab amount depends on current column */
				nextra = (P(P_TS) - 1) - col%P(P_TS);
				for (n = 0; n < nextra; n++)
					extra[n] = ' ';
				c = ' ';
			} else if ( c == NUL && P(P_LS) ) {
				extra[0] = NUL;
				nextra = 1;
				c = '$';
			} else if ( c != NUL && (n=chars[c].ch_size) > 1 ) {
				uchar	*p;

				nextra = 0;
				p = (uchar *)chars[c].ch_str;
				/* copy 'ch-str'ing into 'extra' in reverse */
				while ( n > 1 )
					extra[nextra++] = p[--n];
				c = p[0];
			}
		}

		if ( c == NUL )
			break;

#ifdef KANJI
		if (col == Columns - 1 && kanji2 == FALSE && isKanji(c))
			outchar(' ');
#endif
		outchar(c);
		col++;
	}
}

#ifndef PC98_VRAM

void
screenclear()
{
	register int	i;

	toutstr(T_ED, Rows);	/* clear the display */

	for (i = 0; i < Rows; i++) {
		*Realscreen[i] = NUL;
		*Nextscreen[i] = NUL;
	}
}

#else /*PC98_VRAM*/

void
screenclear()
{
	register int	pos, endpos;

	pos = 0;
	endpos = Rows * Columns;
	while (pos < endpos) {
		vram[pos] = ' ';
		*(unsigned char *)&vattr[pos++] = 0xe1;
	}
}

#endif /*PC98_VRAM*/

void
cursupdate()
{
	register LPTR	*p;
	register int	icnt, c, nlines;
	register int	i;
	int	didinc;
#ifdef KANJI
	int	forcedwrap;	/* wrap is forced (kanji on last column) */
#endif

	if (bufempty()) {		/* special case - file is empty */
		*Topchar  = *Filemem;
		*Curschar = *Filemem;
	} else if ( LINEOF(Curschar) < LINEOF(Topchar) ) {
		nlines = cntllines(Curschar,Topchar);
		/* if the cursor is above the top of */
		/* the screen, put it at the top of the screen.. */
		*Topchar = *Curschar;
		Topchar->index = 0;
		/* ... and, if we weren't very close to begin with, */
		/* we scroll so that the line is close to the middle. */
		if ( nlines > Rows/3 ) {
			for (i=0, p = Topchar; i < Rows/3 ;i++, *Topchar = *p)
				if ((p = prevline(p)) == NULL)
					break;
		} else
			s_ins(0, nlines-1);
		updatescreen();
	}
	else if (LINEOF(Curschar) >= LINEOF(Botchar)) {
		nlines = cntllines(Botchar,Curschar);
		/* If the cursor is off the bottom of the screen, */
		/* put it at the top of the screen.. */
		/* ... and back up */
		if ( nlines > Rows/3 ) {
			p = Curschar;
			for (i=0; i < (2*Rows)/3 ;i++)
				if ((p = prevline(p)) == NULL)
					break;
			*Topchar = *p;
		} else {
			scrollup(nlines);
		}
		updatescreen();
	}

	Cursrow = Curscol = Cursvcol = 0;
	for ( p=Topchar; p->linep != Curschar->linep ;p = nextline(p) )
		Cursrow += plines(p);

	Cline_row = Cursrow;
	Cline_size = plines(p);

	if (P(P_NU))
		Curscol = 8;

	for (i=0; i <= Curschar->index ;i++) {
		c = Curschar->linep->s[i];
#ifdef KANJI
		forcedwrap = FALSE;
		if (isKanji(c)) {
			i++;
			if (Curscol == Columns - 1) {
				Curscol++;
				Cursvcol++;
				forcedwrap = TRUE;
			}
			icnt = 2;
		} else
#endif
		/* A tab gets expanded, depending on the current column */
		if ( c == TAB && !P(P_LS) )
			icnt = P(P_TS) - (Cursvcol % P(P_TS));
		else
			icnt = chars[(unsigned)(c & 0xff)].ch_size;
		Curscol += icnt;
		Cursvcol += icnt;
		if ( Curscol >= Columns ) {
			Curscol -= Columns;
			Cursrow++;
			didinc = TRUE;
		}
		else
			didinc = FALSE;
	}
#ifdef KANJI
	if (didinc && forcedwrap == FALSE)
		Cursrow--;
#else
	if (didinc)
		Cursrow--;
#endif

	if (c == TAB && State == NORMAL && !P(P_LS)) {
		Curscol--;
		Cursvcol--;
	} else {
		Curscol -= icnt;
		Cursvcol -= icnt;
	}
	if (Curscol < 0)
		Curscol += Columns;

	if (set_want_col) {
		Curswant = Cursvcol;
		set_want_col = FALSE;
	}
}

/*
 * The rest of the routines in this file perform screen manipulations.
 * The given operation is performed physically on the screen. The
 * corresponding change is also made to the internal screen image.
 * In this way, the editor anticipates the effect of editing changes
 * on the appearance of the screen. That way, when we call screenupdate
 * a complete redraw isn't usually necessary. Another advantage is that
 * we can keep adding code to anticipate screen changes, and in the
 * meantime, everything still works.
 */

#ifndef PC98_VRAM

/*
 * s_ins(row, nlines) - insert 'nlines' lines at 'row'
 */
void
s_ins(row, nlines)
int	row;
int	nlines;
{
	register int	i;
	char outbuf[20];

	if (nlines <= 0)
		return;

#ifdef TERMCAP
	if (T_IL == NULL && (!T_CS || !T_SR))	/* can't do it */
		return;
#else
	if (T_IL == NULL)			/* can't do it */
		return;
#endif

	/*
	 * It "looks" better if we do all the inserts at once
	 */
	if (T_CI) toutstr(T_CI, 1);	/* disable cursor */
	if (T_SC) toutstr(T_SC, 1);	/* save position */

#ifdef TERMCAP
	if (T_IL) {
		windgoto(row, 0);
		if (nlines > 1 && T_IM) {
#ifdef TERMINFO
			toutstr(tparm(T_IM, nlines), Rows - row);
#else
			sprintf(outbuf, T_IM, nlines);
			toutstr(outbuf, Rows - row);
#endif
		} else {
			for (i=0; i < nlines ;i++)
				toutstr(T_IL, Rows - row);
		}
	} else {
		toutstr(tgoto(T_CS, Rows-2, row), 1);
		windgoto(row, 0);
		for (i = 0; i < nlines; i++)
			toutstr(T_SR, 1);
		toutstr(tgoto(T_CS, Rows-1, 0), 1);
	}
#else
	windgoto(row, 0);
	if (nlines > 1 && T_IM) {
		sprintf(outbuf, T_IM, nlines);
		toutstr(outbuf, Rows - row);
	} else {
		for (i=0; i < nlines ;i++)
			toutstr(T_IL, Rows - row);
	}
#endif

	windgoto(Rows-1, 0);	/* delete any garbage that may have */
	toutstr(T_EL, 1);	/* been shifted to the bottom line */

	if (P(P_MO) && (State == INSERT || State == REPLACE))
		mmsg(State == INSERT? "Insert Mode": "Replace Mode");

	if (T_RC) toutstr(T_RC, 1);	/* restore the cursor position */
	if (T_CV) toutstr(T_CV, 1);	/* enable cursor again */

	/*
	 * Clear the lines to be inserted
	 */
	for (i = 0; i <= nlines; i++)
		*Realscreen[Rows-1-i] = NUL;

	/*
	 * Now do a block move to update the internal screen image
	 */
	for (i = 0; i < Rows-row; i++)
		Realscreen[Rows-1+nlines-i] = Realscreen[Rows-1-i];
	for (i = 0; i < nlines; i++)
		Realscreen[row+i] = Realscreen[Rows+i];
		Realscreen[Rows+i] = NULL;
}

/*
 * s_del(row, nlines) - delete 'nlines' lines at 'row'
 */
void
s_del(row, nlines)
int	row;
int	nlines;
{
	register int	i;
	char outbuf[20];

	if (nlines <= 0)
		return;

#ifdef TERMCAP
	if (T_DL == NULL && T_CS == NULL)	/* can't do it */
		return;
#else
	if (T_DL == NULL)			/* can't do it */
		return;
#endif

	/* delete the lines */
	if (T_CI) toutstr(T_CI, 1);		/* disable cursor */
	if (T_SC) toutstr(T_SC, 1);		/* save position */

#ifdef TERMCAP
	if (T_DL) {
		windgoto(row, 0);
		if (nlines > 1 && T_DM) {
#ifdef TERMINFO
			toutstr(tparm(T_DM, nlines), Rows - row);
#else
			sprintf(outbuf, T_DM, nlines);
			toutstr(outbuf, Rows - row);
#endif
			windgoto(Rows-1-nlines, 0);
			toutstr(T_EL, 1);
		} else {
			for (i=0; i < nlines ;i++) {
				toutstr(T_DL, Rows - row);
				if (i != 0)
					continue;
				windgoto(Rows-2, 0);
				toutstr(T_EL, 1);
				windgoto(row, 0);
			}
		}
	} else {
		toutstr(tgoto(T_CS, Rows-2, row), 1);
		windgoto(Rows-2, 0);
		for (i = 0; i < nlines; i++)
			toutstr(T_SF, 1);
		toutstr(tgoto(T_CS, Rows-1, 0), 1);
		windgoto(Rows-1, 0);		/* delete any garbage that */
		toutstr(T_EL, 1);		/* was on the status line */
	}
#else
	windgoto(row, 0);
	if (nlines > 1 && T_DM) {
		sprintf(outbuf, T_DM, nlines);
		toutstr(outbuf, Rows - row);
		windgoto(Rows-1-nlines, 0);	/* delete any garbage that */
		toutstr(T_EL, 1);		/* was on the status line */
	} else {
		for (i=0; i < nlines ;i++) {
			toutstr(T_DL, Rows - row);   /* delete a line */
			if (i != 0)
				continue;
			windgoto(Rows-2, 0);	/* delete any garbage that */
			toutstr(T_EL, 1);	/* was on the status line */
			windgoto(row, 0);
		}
	}
#endif

	if (P(P_MO) && (State == INSERT || State == REPLACE))
		mmsg(State == INSERT? "Insert Mode": "Replace Mode");

	if (T_RC) toutstr(T_RC, 1);		/* restore position */
	if (T_CV) toutstr(T_CV, 1);		/* enable cursor again */

	/*
	 * Clear the lines to be inserted
	 */
	for (i = 0; i < nlines; i++)
		*Realscreen[row+i] = NUL;
	*Realscreen[Rows-1] = NUL;

	/*
	 * do a block move to update the internal image
	 */
	for (i = 0; i < nlines; i++)
		Realscreen[Rows+i] = Realscreen[row+i];
	for (i = 0; i < Rows-row; i++)
		Realscreen[row+i] = Realscreen[row+nlines+i];
	for (i = 0; i < nlines; i++)
		Realscreen[Rows+i] = NULL;
}

#else /*PC98_VRAM*/

/*
 * s_ins(row, nlines) - insert 'nlines' lines at 'row'
 */
void
s_ins(row, nlines)
int	row;
int	nlines;
{
	register int	i, j;

	/*
	 * if (nlines <= 0)
	 *	return;
	 *
	 * i = (Rows - 1) * Columns;
	 * j = (Rows - nlines - 1) * Columns;
	 * while (j > row * Columns)
	 *	vram[--i] = vram[--j];
	 * i = row * Columns;
	 * while (i < (row + nlines) * Columns)
	 *	vram[i++] = ' ';
	 */

	/* clear the bottom line */
	i = (j = Rows * Columns) - Columns;
	while (i < j) {
		vram[i] = ' ';
		*(uchar far *)&vattr[i++] = A_NORM;
	}
	if (P(P_MO) && (State == INSERT || State == REPLACE))
		mmsg(State == INSERT? "Insert Mode": "Replace Mode");
}

/*
 * s_del(row, nlines) - delete 'nlines' lines at 'row'
 */
void
s_del(row, nlines)
int	row;
int	nlines;
{
	register int	i, j;

	/*
	 * if (nlines <= 0)
	 *	return;
	 *
	 * i = row * Columns;
	 * j = (row + nlines) * Columns;
	 * while (j < (Rows - 1) * Columns)
	 *	vram[i++] = vram[j++];
	 * i = (Rows - nlines - 1) * Columns;
	 * while (i < (Rows - 1) * Columns)
	 *	vram[i++] = ' ';
	 */

	/* clear the bottom line */
	i = (j = Rows * Columns) - Columns;
	while (i < j) {
		vram[i] = ' ';
		*(uchar far *)&vattr[i++] = A_NORM;
	}
	if (P(P_MO) && (State == INSERT || State == REPLACE))
		mmsg(State == INSERT? "Insert Mode": "Replace Mode");
}

#endif /*PC98_VRAM*/
