/*
 * $Header: /nw/tony/src/stevie/src/RCS/stevie.h,v 1.19 89/07/12 21:33:32 tony Exp $
 *
 * Main header file included by all source files.
 */

#include "env.h"	/* defines to establish the compile-time environment */

#ifdef KANJI
typedef unsigned char	uchar;
#else
#define	uchar	char
#endif

#include <stdio.h>
#include <ctype.h>

#ifdef	BSD
#include <strings.h>
#define strchr index
#else
#ifdef	MINIX
extern	char	*strchr();
extern	char	*strrchr();
extern	char	*strcpy();
extern	char	*strcat();
extern	int	strlen();
#else
#include <string.h>
#endif
#endif

#include "ascii.h"
#include "keymap.h"
#include "param.h"
#include "term.h"

/*
 * maximum amount of space for each object (can be modified).
 */
#define	YBSIZE	8192		/* size of yank buffer */
#define	OBSIZE	2048		/* size of output buffer */
#ifdef PC98_MOUSE
#define	IBSIZE	2048		/* size of insertion/paste buffer */
#else
#define	IBSIZE	1024		/* size of insertion buffer */
#endif
#define	GBSIZE	1024		/* size of :n globbing buffer */
#define	MBSIZE	 512		/* size of mapc buffer */
#define	MAXLINE	 512		/* max length of a line */
#define	MAXTLIN	 256		/* max length of a line in tags file */
#define	MAXCLIN	 128		/* max length of a command line */
#define	MAXPLEN	 128		/* max length of a path name */
#define	MAXMLEN	 128		/* max length of a message */
#define	MAXMAP	 128		/* max number of map */

#ifdef PC98_MOUSE
#define	YANK_MAX	1920	/* max length of cut/paste */
#endif

/*
 * editor state
 */
#define NORMAL	 0
#define CMDLINE  1
#define INSERT   2
#define REPLACE  3
#define FORWARD  4
#define BACKWARD 5

/*
 * Boolean type definition and constants
 */
typedef	short	bool_t;

#ifndef	TRUE
#define	FALSE	(0)
#define	TRUE	(1)
#endif

/*
 * SLOP is the amount of extra space we get for text on a line during
 * editing operations that need more space. This keeps us from calling
 * malloc every time we get a character during insert mode. No extra
 * space is allocated when the file is initially read.
 */
#define	SLOP	10

/*
 * LINEINC is the gap we leave between the artificial line numbers. This
 * helps to avoid renumbering all the lines every time a new line is
 * inserted.
 */
#define	LINEINC	10

#define CHANGED		Changed=TRUE
#define UNCHANGED	Changed=FALSE

struct	line {
	struct	line	*prev, *next;	/* previous and next lines */
	uchar	*s;			/* text for this line */
	int	size;			/* actual size of space at 's' */
	unsigned long	num;		/* line "number" */
};

#define	LINEOF(x)	((x)->linep->num)

struct	lptr {
	struct	line	*linep;		/* line we're referencing */
	int	index;			/* position within that line */
};

typedef	struct line	LINE;
typedef	struct lptr	LPTR;

struct charinfo {
	char ch_size;
	char *ch_str;
};

extern struct charinfo chars[];

/*
 * structure for :map, :map!, :ab command
 */
struct map {
	int	m_type;			/* type of map */
	int	m_len;			/* length of key sequence */
	char	*m_name;		/* name for key sequence */
	char	*m_key;			/* key sequence */
	char	*m_str;			/* string to be mapped */
};
typedef struct map	MAP;

/*
 * map type
 */
#define	MAPC	0			/* :map */
#define MAPI	1			/* :map! */
#define	ABBR	2			/* :ab */

#ifdef KANJI
#ifdef BIT8
#define	isKanji(c)	FALSE
#else
extern	uchar		kanji1[];
#define	isKanji(c)	kanji1[c]
#endif
#endif

/*
 * file access status
 */
#define	FS_EXIST	0x01		/* file exists */
#define	FS_READ		0x02		/* file is readable */
#define	FS_WRITE	0x04		/* file is writable */
#define	FS_DIR		0x08		/* file is a directory */

extern	int	State;
extern	int	Rows;
extern	int	Columns;
extern	uchar	**Realscreen;
extern	uchar	**Nextscreen;
extern	uchar	*Filename;
extern	LPTR	*Filemem;
extern	LPTR	*Filetop;
extern	LPTR	*Fileend;
extern	LPTR	*Topchar;
extern	LPTR	*Botchar;
extern	LPTR	*Curschar;
extern	LPTR	*Insstart;
extern	int	Cursrow, Curscol, Cursvcol, Curswant;
extern	bool_t	set_want_col;
extern	int	Prenum;
extern	bool_t	Changed;
extern	uchar	Redobuff[], Insbuff[];
extern	uchar	*Insptr;
extern	int	Ninsert;
extern	bool_t	got_int;
extern	MAP	Maptab[];
extern	int	nmap;
extern	bool_t	can_ab;
extern	uchar	*Initfile;	/* given a name in system-dependent files */

extern	char	*malloc();

/*
 * alloc.c
 */
uchar	*alloc(), *strsave(), *mkstr();
void	screenalloc(), filealloc(), freeall();
LINE	*newline();
bool_t	bufempty(), buf1line(), lineempty(), endofline(), canincrease();

/*
 * cmdline.c
 */
uchar	*pretty();
void	docmdln(), dotag(), msg(), emsg(), smsg(), gotocmd(), wait_return();
uchar	*getcmdln();
void	doinitfile();
void	domap();

/*
 * edit.c
 */
void	edit(), insertchar(), getout(), scrollup(), scrolldown(), beginline();
bool_t	deletechar(), oneright(), oneleft(), oneup(), onedown();

/*
 * fileio.c
 */
void	filemess(), renum();
bool_t	readfile(), writeit();

/*
 * help.c
 */
bool_t	help();

/*
 * linefunc.c
 */
LPTR	*nextline(), *prevline(), *coladvance();

/*
 * main.c
 */
void	stuffin(), stuffnum();
void	do_mlines();
int	vgetc();
bool_t	anyinput();

/*
 * mark.c
 */
void	setpcmark(), clrall(), clrmark();
bool_t	setmark();
LPTR	*getmark();

/*
 * misccmds.c
 */
void	opencmd(), fileinfo(), inschar(), delline();
bool_t	delchar();
int	cntllines(), plines();
LPTR	*gotoline();

/*
 * normal.c
 */
void	normal();

/*
 * param.c
 */
void	doset();

/*
 * ptrfunc.c
 */
int	inc(), dec();
#ifdef KANJI
int	kinc(), kdec();
#endif
int	gchar();
void	pchar(), pswap();
bool_t	lt(), equal(), ltoreq();
#if 0
/* not currently used */
bool_t	gtoreq(), gt();
#endif

/*
 * screen.c
 */
void	updatescreen(), updateline();
void	screenclear(), cursupdate();
void	s_ins(), s_del();
void	prt_line();

/*
 * search.c
 */
int	dosub();
void	doglob();
bool_t	searchc(), crepsearch(), findfunc(), dosearch(), repsearch();
LPTR	*showmatch();
LPTR	*fwd_word(), *bck_word(), *end_word();

/*
 * undo.c
 */
bool_t	u_save(), u_saveline();
void	u_clear();
void	u_lcheck(), u_lundo();
void	u_undo();

/*
 * Machine-dependent routines.
 */
int	inchar();
void	flushbuf();
void	outchar(), outstr(), beep();
void	tstpon(), tstpoff();
void	windinit(), windexit(), windgoto();
void	delay();
FILE	*fopenb();
int	filestat();
int	Rename();
uchar	*backupname();
uchar	*fixname();
uchar	*mkpath();
void	doshell(), cmdread();

#ifdef FEPCTRL
/*
 * fepctrl.c
 */
int	fep_init();
void	fep_on(), fep_off(), fep_force_off(), fep_term();
#endif
